from __future__ import annotations
from typing import List,Dict
import curses
import math
import fnmatch

class KeyboardInput:
	code:int
	c:str
	name:str

	def __init__(self,code:int):
		self.code = code
		self.c = chr(code)
		self.name = codeToName.get(code,None)

	def __repr__(self):
		return f'Code : {self.code} c : {self.c} name : {self.name}'
class Curses:
	def __init__(self,stdscr):
		self.stdscr : curses.window = stdscr
		curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
		curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_CYAN)

		self.colors_normal = curses.color_pair(1)
		self.colors_selected = curses.color_pair(2)

	def clear(self):
		self.stdscr.erase()
	
	def emptyLine(self,n=1):
		self.stdscr.addstr("\n"*n)
	
	def line(self,txt:str='',selected=False,color=None,underline=False,bold=False,prefix='',add_line_break=True):
		if color is None:
			color = self.colors_selected if selected else self.colors_normal
		attr = color

		for should_do,attrValue in [
			[underline,curses.A_UNDERLINE],
			[bold,curses.A_BOLD]]:
			if should_do:
				attr = attr | attrValue
				

		
		if prefix:
			self.stdscr.addstr(prefix)
		
		self.stdscr.addstr(txt,attr)
		if add_line_break:
			self.stdscr.addstr("\n")


	def wait_for_input(self):
		code = self.stdscr.getch()
		return KeyboardInput(code)
	

def to_list(x=None,default=None):
	if x is None:
		return default
	
	return x if isinstance(x,list) else [x]


def _getcurses_codes():

	extra_keys = dict(
		enter=[10,459],
		esc=[27],
		backspace=[8]
	)


	codeToName = {}
	for k,v in curses.__dict__.items():
		if not k.lower().startswith('key_'):
			continue
		codeToName[int(v)] = k.lower()[4:]


	for keyname,codes in extra_keys.items():
		for code in codes:
			codeToName[code] = keyname
	
	
	
	

	nameToCode = dict()
	for k,v in codeToName.items():
		if not v in nameToCode:
			nameToCode[v] = []
		nameToCode[v].append(k)
	return codeToName,nameToCode


codeToName, nameToCode = _getcurses_codes()



def _get_keyname(keyname,quit=False):
	if keyname in codeToName:
		keyname = codeToName[keyname]


	if keyname in nameToCode:
		return keyname
	
	if quit:
		print(f'Invalid keyname {keyname}')
		exit()

	
	

class Command:
	def __init__(self,name,func=None,desc=None,cmd_keys=None,exit_after:bool=False,with_parameters=False,data=None):
		self.name 			:str		= name
		self.with_parameters = with_parameters
		self.func 						= func
		self.desc 			:str		= desc if desc is not None else name
		self.cmd_keys	  	:List[str]  = to_list(cmd_keys,[])
		self.exit_after 				= exit_after
		self.data						= data

	def exec(self,*a):
		if not self.func:
			print(f'{self}.exec func is none')
			return
		if self.with_parameters:
			self.func(*a)
		else:
			self.func()

		if self.exit_after:
			exit()

	def __repr__(self):
		return f'Command : {self.name}'

	
class ItemToPrint:
	def __init__(self,row_number:int,result_number:int,selected:bool,cmd:Command):
		self.row_number = row_number
		self.result_number = result_number
		self.selected = selected
		self.cmd = cmd

def search_filter_default(c:Command,search_input):
	return c.name.lower().find(search_input.lower()) != -1


class SearchFilterHelper:
	@classmethod
	def fnmatch(cls):
		
		def ret_val(c:Command,search_input:str):
			return fnmatch.fnmatch(c.name,search_input)
		
		return ret_val

	@classmethod
	def default(cls):
		def ret_val(c:Command,search_input:str):
			return fnmatch.fnmatch(c.name,'*' + search_input + '*')
		
		return ret_val
class SearchFilter:
	fnmatch = SearchFilterHelper.fnmatch()
	default = SearchFilterHelper.default()

class Commands(list):
	def __init__(self, *commands:List[Command],max_rows:int=20):

		self.search_filter = search_filter_default

		self.clear()
		self.max_rows = max_rows
		# self.__selected_index = 0
		# self.item_start_index = 0
		# self.max_rows = max_rows
		# self.commandline_keys = dict()
		# self.append(*commands)
		# self.search_input = ''
		# self._searching = False
		self.append(*commands)

	def clear(self):
		self.commandline_keys = dict()
		self.__selected_index = 0
		self.item_start_index = 0
		self.search_input = ''
		self._searching = False

		super().clear()


	def load_key_listeners(self,menu:Menu):
		@menu.key(curses.KEY_BACKSPACE)
		def back_space(m:Menu):
			print(m.cmds.searching)
			if not m.cmds.searching:
				return
			m.cmds.search_input = m.cmds.search_input[:-1]
	@property
	def searching(self):
		return self._searching

	@searching.setter
	def searching(self,searching):
		if searching == self.searching:
			return
		
		print(f'Set searching : {searching}')
		self._searching = searching

		self.search_input = ''
		



	def __check_item(self,c:Command):
		for sc in c.cmd_keys:
			if sc in self.commandline_keys:
				print(f'{sc} {c} already in cmd_keys !')
				continue
			self.commandline_keys[sc] = c
		return True
	
	def append(self,*ar:List[Command]):
		for v in ar:
			if not self.__check_item(v):
				continue
			super().append(v)
		
	def __getitem__(self, index)->Command:
		return super().__getitem__(index)
	
	def __setitem__(self, key, v:Command):
		if not self.__check_item(v):
			return
		return super().__setitem__(key, v)




	def set_items_filtered(self):
		def get() -> List[Command]:
			if not self.searching:
				return self
			if not self.search_input:
				return self


			
			def search_filter(c:Command):
				return self.search_filter(c,self.search_input)

			# def search_filter_and(x:Command):
			# 	nl = x.name.lower()
			# 	for a in self.search_input.lower():
			# 		if a and (nl.find(a) == -1):
			# 			return False
			# 	return True
			
			# def search_filter(x:Command):
			# 	return x.name.lower().find(self.search_input.lower()) != -1
			
			return list(filter(search_filter,self))

		self.items_filtered = get()
	



	@property
	def items_to_print(self) -> List[ItemToPrint]:
		self.set_items_filtered()

		row_number = 0
		for result_number in range(self.first_index,self.last_index+1):
			row_number += 1
			cmd_selected = result_number == self.__selected_index
			cmd = self.items_filtered[result_number]
			
			yield ItemToPrint(row_number=row_number,result_number=result_number,selected=cmd_selected,cmd=cmd)
			#yield row_number,cmd_index,cmd_selected,cmd
		


	@property
	def first_index(self):
		return max(0,self.item_start_index)
	
	@property
	def last_index(self):
		return min(len(self.items_filtered) - 1, self.first_index + self.max_rows - 1)


	@property
	def selected_index(self):
		return self.__selected_index

	@selected_index.setter
	def selected_index(self,new_value):
		last_value = self.__selected_index
		new_value  = min(max(0,new_value ), len(self.items_filtered) - 1)
		delta_selected_index = (new_value - last_value)

		if delta_selected_index == 0:
			return
		elif delta_selected_index == 1:
			#one down
			if (self.last_index == last_value):
				self.item_start_index += 1
		elif delta_selected_index == -1:
			#one up
			if (self.first_index == last_value):
				self.item_start_index -= 1
		else:

			self.item_start_index = min(max(0,new_value - math.floor(self.max_rows/2)),max(0,len(self.items_filtered) - self.max_rows))
		
		self.__selected_index = new_value

	@property
	def selected_cmd(self) -> Command:
		try:
			x = self.items_filtered[self.selected_index]
			return x
		except:
			return None
		
class Menu:
	title:str=''
	def __init__(self,*commands : List[Command],start=False,title=None,max_rows=20):
		
		self.title = title
		self.__on_key = dict()
		self.cmds  = Commands(*commands,max_rows=max_rows)



		@self.key('f2')
		def search(menu:Menu):
			menu.cmds.searching = not menu.cmds.searching
			

		
		@self.key('up')
		def up(menu:Menu):
			menu.cmds.selected_index -= 1
			
		@self.key('ppage')
		def up(menu:Menu):
			menu.cmds.selected_index -= menu.cmds.max_rows

		@self.key('npage')
		def down(menu:Menu):
			menu.cmds.selected_index += menu.cmds.max_rows

		@self.key('down')
		def down(menu:Menu):
			menu.cmds.selected_index += 1

		@self.key('enter')
		def enter(menu:Menu):
			if not (cmd := menu.cmds.selected_cmd):
				return
			cmd.exec(menu,cmd)



		

		if start:
			self.start()



	
	def key(self,keyname):
		keyname = _get_keyname(keyname,quit=True)
		
		def decorator(f):
			self.__on_key[keyname] = f
			return f
		return decorator
	


	def do_key(self,keyname):
		keyname = _get_keyname(keyname)
		if not keyname:
			return
		
		if keyname in self.__on_key:
			self.__on_key[keyname](self)
		
	def start(self):
		self.cmds.load_key_listeners(menu=self)
		def wrapper(stdscr):
			C = Curses(stdscr)
			def printer():
				C.clear()

				if self.title:
					C.line(self.title,underline=True)
					C.emptyLine(2)

				for item in self.cmds.items_to_print:
					C.line(item.cmd.name,selected=item.selected,prefix=f'{str(item.result_number).zfill(2)} {str(item.row_number).zfill(2)}   ')
					
				if self.cmds.searching:
					C.emptyLine(2)
					C.line(f'#results : {len(self.cmds.items_filtered)}')
					C.line(f'Searching : {self.cmds.search_input}',add_line_break=False)
					

				userInput = C.wait_for_input()

				if self.cmds.searching and (userInput.c) and (not userInput.name):
					self.cmds.search_input = self.cmds.search_input + userInput.c
					
				else:

					self.do_key(userInput.name)
					
					
					
					

			while True:
				printer()

		
		curses.wrapper(wrapper)

	
	def command(self,name,desc=None,cmd_keys=None,exit_after=False,with_parameters=False,data=None):
		def decorator(f):
			self.cmds.append(Command(name,f,desc=desc,cmd_keys=cmd_keys,exit_after=exit_after,with_parameters=with_parameters,data=data))
			return f
		return decorator

	



from jivi.oe.Imports import *
exec(Wrapper.main_setter)


class Session(Wrapper):
	name = "Session"
	todo = ['load_available_char','load_all_files','load_all_files_tab','load_tables','set_replacers']
	
	reload = Data(
					available_chars = 0,
					all_files 		= 0,
					tables			= 0,
					all_files_tab 	= 0
				)
	def __init__(self,MAIN):
		set_main(MAIN)
		
		self.propath = ENV.propath
		
		self.available_chars = []
		FP.available_chars = FP.data + "available_chars.json"
		
		self.all_files = {}
		self.all_files_tab = {}
		FP.all_files = FP.data + "all_files.json"
		FP.all_files_tab = FP.data + "all_files_tab.json"
		FP.tables = FP.data + "tables.json"
		self.tables = {}
		
		self.do(allinfo=0)
		self.info('loaded')

		
		# self.load_t_f(tabs['file'],tabs['field'])
		
	def load_all_files_tab(self):
		if not self.reload.all_files_tab:
			self.all_files_tab = File.jread(FP.all_files_tab)
			
		if self.all_files_tab: return self.all_files_tab
		
		self.all_files_tab = {fp : (i + 1) for i,fp in enumerate(self.all_files)}
		File.jwrite(FP.all_files_tab,self.all_files_tab)
		return self.all_files_tab
	def load_all_files(self):
		if not self.reload.all_files:
			self.all_files = File.jread(FP.all_files)

		if self.all_files: return self.all_files
		t = {}
		for d in self.propath:
			for fp in Dir.files(d,ex=['w','i','p'],full=1,deep=1):
				t[fp.lower().strip()] = 1
		self.all_files = sorted(list(t.keys()))

		
		File.jwrite(FP.all_files,self.all_files)
		return self.all_files
	
	def set_replacers(self):
		for i,a in enumerate(CONSTANTS):
			exec("REPLACER.{a} = self.available_chars[{i}]".format(a=a,i=i))
	
	def load_available_char(self):
		if not self.reload.available_chars:
			self.available_chars = File.jread(FP.available_chars)
		if self.available_chars: return self.available_chars
		self.load_all_files()
		self.available_chars = []
		self.todo_files = list(Ar.steps(self.all_files,50))
		self.todo = len(self.todo_files)
		self.good_chars = []
		self.good_chars_response = []
		for i in range(255):
			if chr(i).lower() == chr(i).upper() and len(chr(i)) and (len(chr(i))) == len(chr(i).strip()):
				self.good_chars.append(chr(i))
				self.good_chars_response.append([1]*self.todo)
		for i in range(self.todo):
			Thread(self.set_available_chars_helper,(i,))
		while self.todo:
			sleep(0.01)
		self.todo = len(self.todo_files)
		for i,char in enumerate(self.good_chars):
			if sum(self.good_chars_response[i]) == self.todo:
				self.available_chars.append(char)
		File.jwrite(FP.available_chars,self.available_chars)
		return self.available_chars

	def set_available_chars_helper(self,i):
		for fp in self.todo_files[i]:
			x = File.all_full(fp)
			if x:
				for j,v in enumerate(self.good_chars):
					if not self.good_chars_response[j][i]: continue
					if x.find(self.good_chars[j]) != -1:
						self.good_chars_response[j][i] = 0
		self.todo -= 1
	

	def fp_compiled(self,fp,first_step=1):
		
		return (FP.compiled_first_step if first_step else FP.compiled) + "{i}.json".format(i=self.all_files_tab[fp])

	def fp_dir(self,fp):
		if not os.path.isdir(fp):
			fp = os.path.join(self.wd,fp)
		if not os.path.isdir(fp):
			return 0
		return os.path.abspath(fp).rstrip(os.sep) + os.sep
		
		

	def load_tables(self):
		if not self.reload.tables:
			self.tables = File.jread(FP.tables)
		if self.tables: return self.tables


		ENV.load_env()
		self.load_t_f(ENV.tabs['file'],ENV.tabs['field'])
		File.jwrite(FP.tables,self.tables)
		return self.tables
		
		
	
		
	def load_t_f(self,tables,fields):
		self.tables = {}
		trecids = {}
		for a in tables:
			name = a['_File-Name'].lower().strip()
			self.tables[name] = dict(tab=a,fields={})
			trecids[a['rrecid']] = name
			
		for field in fields:
			
			name = field['_Field-Name'].lower().strip()
			tablename = trecids[field['_File-recid']]
			self.tables[tablename]['fields'][name] = field
			
	def get_table(self,tn):
		tn = tn.lower()
		if tn in self.tables:
			return self.tables[tn]
		
	def get_field(self,tn,fn=0):
		if not fn:
			ar = tn.split('.')
			if len(ar) == 2:
				return self.get_field(*ar)
			return 0
		table = self.get_table(tn)
		if not table:
			return 0
		fn = fn.lower().strip()
		if fn in table['fields']:
			return table['fields'][fn]
			
	def fp(self,fp):
		if os.path.isfile(fp):
			return os.path.abspath(fp)
		for d in self.propath:
			x = os.path.join(d,fp)
			if os.path.isfile(x):
				return os.path.abspath(x)
		return 0


	

from .Viewas import viewas

class HasHelper:
	def __init__(self,ob):
		self.ob = ob
		
	def __getattr__(self,k):
		ob = getattr(self,'ob')
		if k == 'ob': return ob
		return k in ob.t
		

class Object:
	attr = []
	attr.append(dict(n='name',r=1,lower=1))

	
	can_set = {}
	name = "unknown"
	error_pause = 0
	def __init__(self,env=0,window=0,tab={}):
		self.error_pause = 0
		self.errors = []
		self.set_attr()
		self.env = env
		self.has = HasHelper(self)
		self.t = {}
		self.window = window
		try:
			self.session = self.env.session
		except:
			pass
		
		if window:
			self.env = self.window.env
		if tab:
			self.set_tab(tab)
		
		self.va = viewas(self)
			
			
	def set_attr(self):
		self.attibute = {}
		self.can_set = {}
		
		default = dict(r=0,s=1,a=0,lower=0)
		classes = type(self).mro()[:-1]
		attr = {}
		self.main_object = classes[0]
		if not hasattr(self.main_object,'ignore'):
			self.main_object.ignore = {}
		
		self.type = classes[0].__name__
		for cl in classes:
			if not hasattr(cl,'attr'): continue
			for t in cl.attr:

				n = t['n'].lower()
				if not n in self.attibute:
					self.attibute[n] = {}
					self.attibute[n].update(default)
					
				if 'a' in t:
					for x in t['a'].split(','):
						x = x.lower()
						if not x: continue
						if x in self.can_set and self.can_set[x] != n:
							self.error('set_attr {x} already set to {old} not {new}'.format(x=x,old=self.attr_alias[x],new=n))
						
						self.can_set[x] = n
				self.can_set[n] = n
				self.attibute[n].update(t)
		
				

		

		
	def set(self,a,b):
		a = a.lower()
		if not a in self.can_set:
			if not a in self.main_object.ignore:
				self.main_object.ignore[a] = 1
			else:
				return
				
			return self.error("cant set " + a)
			
		attr = self.attibute[self.can_set[a]]
		if attr['lower']:
			b = b.lower().strip()
		setattr(self,attr['n'],b)
		self.t[attr['n']] = b
		
	def set_tab(self,tab={}):

				
		self.error_pause = 1
		for a,b in tab.items():
			self.set(a,b)
		
		self.error_resume()
		
	def error_resume(self):
		self.error_pause = 0
		if not self.errors: return
		self.error(" ".join(self.errors))
		self.errors = []
	def html(self):
		return "no html"
		
		
	def error(self,s):
		if self.error_pause:
			self.errors.append(s)
			return
		print("[{type}]{name} : {s}".format(type=self.type,name=self.name,s=s))
	
	
	def __setattr__(self,k,v):
		if not k in self.can_set:
			return object.__setattr__(self,k,v)
		self.t[self.attibute[self.can_set[k]]['n']] = v
		return object.__setattr__(self,self.attibute[self.can_set[k]]['n'],v)
	
	
	def htp(self,*a,**b):
		return self.session.htp(*a,**b)
		
	def wtp(self,*a,**b):
		return self.session.wtp(*a,**b)

	def ctp(self,*a,**b):
		return self.session.ctp(*a,**b)

	def rtp(self,*a,**b):
		return self.session.rtp(*a,**b)
		
	def ptc(self,*a,**b): 		return self.session.ptc(*a,**b)


"""
	comments in preproc => verwijder
	comments in strings => NIET !
	


"""

from jivi.oe.constants import *
from jivi.fs import *

from StripperGood import *
ALL_KW = {k : 1 for k in ALIAS.keys()}
for k in DEFINE_TYPES.keys(): ALL_KW[k] = 1


def is_kw(s):
	s = s.lower().strip()
	return s in ALL_KW
from time import time

strip_string_comments_preproc_ar = [(a,b,len(a)) for a,b in [('/*',1),('*/',-1),('"',2),("'",3),('{',5),('}',-5),('&glob',6),('&scop',7)]]
"""
eerst alle strings en comments weg daarna dubbele spaties
"""

is_number_val = 8


is_string = {3 : 1,2 : 1}
is_comment = {1 : 1, -1 : 1}
is_include = {5 : 1, -5 : 1}
is_scope = {6 : 1, 7 : 1}
string_appenders = {':%s' % k : 1 for k in 'r,l,c,t,u'.split(',')}
str_numbers = {str(i) : 1 for i in range(10)}
is_scope_local = 7


def is_decimal(x):
	try:
		s = float(x)
		return 1
	except:
		return 0
		
def should_replace(s):
	if s == '.': return 0
	if is_decimal('0' + s): return 1
	s = s.strip()
	ind = s.find('.',1) 
	if ind != -1 and ind != (len(s) - 1): return 1
	


class Stripper(StripperGood):
	dos = ['read_file','set_replace_char','strip_string_coment','do_preproc','remove_preproc_conditions','strip_rest','create_commands','create_indent']
	dos = ['read_file','set_replace_char','strip_string_coment']
	

		
	def strip_string_coment(self):
		ignore_indexes = {}
		index = self.txt.find('~')
		while index != -1:
			ignore_indexes[index+1] = 1
			index = self.txt.find('~',index+1)
	
	
	def remove_preproc_conditions(self):
		index_now = self.txt.find('&')
		while index_now != -1:
			stop_index = self.txt.find(" ",index_now)
			txt = self.txt[index_now:stop_index]
			print(txt)
			exit()
			shouldbreak = 0
			if stop_index == -1: 
				stop_index = len(self.txt)
				shouldbreak = 1
			
			r = self.txt[index_now:stop_index].rstrip()


		
	def strip_string_comments_preproc(self):
		ignore_indexes = {}
		index = self.txt.find('~')
		while index != -1:
			ignore_indexes[index+1] = 1
			index = self.txt.find('~',index+1)
		
		
		
		count = 0

		ar = []
		for needle,value,la in strip_string_comments_preproc_ar:
			index = self.txt.find(needle)
			while index != -1:
				ar.append((index,value))
				index = self.txt.find(needle,index+la)
		
		
		ar.sort()
		inum = 0
		lar = len(ar)
		larminone = lar - 1
		temp_ar = []
		
		add_to_replaced = 0
		start_index_replaced = 0
		stop_index_replaced = 0
		last_stop = 0
		while inum < len(ar):
			
			start_inum = inum
			start_index_replaced,val = ar[inum]
			stop_index_replaced = 0
			if val in is_scope:
				stop_index_replaced = self.txt.find("\n",start_index_replaced)
				shouldbreak = 0
				if stop_index_replaced == -1: 
					stop_index_replaced = len(self.txt)
					shouldbreak = 1
				
				r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
					
						
				while r.endswith('~'):
					r = r[:-1]
					if shouldbreak:
						break
					stop_index_replaced = self.txt.find("\n",stop_index_replaced + 1)
					shouldbreak = 0
					if stop_index_replaced == -1: 
						stop_index_replaced = len(self.txt)
						shouldbreak = 1
					r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
				
				add_to_replaced = 0
				self.scopes.append(((val == is_scope_local),r))
		
				
			elif val in is_string:
				add_to_replaced = 1
				inum += 1
				string_val = val
				string_found = 0
				while inum < larminone:
					if (ar[inum][1] != string_val) or (ar[inum][0] in ignore_indexes):
						inum += 1
					elif (ar[inum + 1][1] == string_val) and (ar[inum + 1][0] == (ar[inum][0] + 1)):
						inum += 2
					else:
						string_found = ar[inum][0]
						break
				if (not string_found) and (inum < lar) and (ar[inum][1] == string_val):
					string_found = ar[inum][0]
				if string_found and self.txt[string_found + 1:string_found + 3].lower() in string_appenders:
					string_found += 2
					
					while self.txt[string_found+1] in str_numbers:
						string_found += 1
				if not string_found:
					self.add_error_at_pos("do_str_c_pre : not string_found",start_index_replaced,func='add_all_strings')
					string_found = start_index_replaced
					inum = start_inum
				
				start = start_index_replaced
				stop_index_replaced = string_found + 1

			elif val in is_include:
				add_to_replaced = 1
				count = val
				inum += 1
				found = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_include:
						count += val
						if count < 0: print("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 1

							found = 1
							break
					inum += 1
				if not found:
					self.add_error_at_pos("do_str_c_pre : is_include not found! started at line ",start_index_replaced)
			elif val in is_comment:
				add_to_replaced = 0
				last_index_plus_one = 0
				count = val
				inum += 1
				found_comment = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_comment and last_index_plus_one != index:
						count += val
						if count < 0: print("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 2
							found_comment = 1
							break
						last_index_plus_one = index + 1
					inum += 1
				if not found_comment:
					self.add_error_at_pos("do_str_c_pre : is_comment not found_comment! started at line ",start_index_replaced)
					
					
			while inum < lar and (ar[inum][0] < stop_index_replaced):
				inum += 1
			if last_stop != start_index_replaced:
				temp_ar.append(self.txt[last_stop:start_index_replaced])
			if add_to_replaced:
				self.replaced.append(self.txt[start_index_replaced:stop_index_replaced])
				temp_ar.append(self.replace_with)
			last_stop = stop_index_replaced
				
		temp_ar.append(self.txt[stop_index_replaced:])
		self.txt = "".join(temp_ar)
		return len(self.errors) == 0
		
	strip_rest_replacers = ['[',']',',','. ','(',')']
	strip_rest_replacers_ends = 'procedure,function,case,for,triggers'.split(',')
	

	
	def strip_rest(self):
		index_now = self.txt.find('&')
		while index_now != -1:

			stop_index = self.txt.find("\n",index_now)
			shouldbreak = 0
			if stop_index == -1: 
				stop_index = len(self.txt)
				shouldbreak = 1
			
			r = self.txt[index_now:stop_index].rstrip()
				
					
			while r.endswith('~'):
				r = r[:-1]
				if shouldbreak:
					break
				stop_index = self.txt.find("\n",stop_index + 1)
				shouldbreak = 0
				if stop_index == -1: 
					stop_index = len(self.txt)
					shouldbreak = 1
				r = self.txt[index_now:stop_index].rstrip()
				
			self.txt = self.txt[:index_now] + self.txt[stop_index:]
			index_now = self.txt.find('&')
		self.txt = self.txt.replace("\t"," ").replace("\n"," ").lower().replace(self.replace_with,' ' + self.replace_with + ' ').replace(':','.')
		for a in self.strip_rest_replacers:
			self.txt = self.txt.replace(a,' ' + a + ' ')

		
		while self.txt.find('  ') != -1:
			self.txt = self.txt.replace('  ',' ')
		for a in self.strip_rest_replacers_ends:
			self.txt = self.txt.replace(' end ' + a,' end ')

		self.txt = self.txt.replace(' forward ',' end ')
		while self.txt.find('  ') != -1:
			self.txt = self.txt.replace('  ',' ')
		self.txt = self.txt.strip()
		return 1

	indent_starts_anywhere = ['do','case','repeat']
	indent_starts = ['function','procedure']
	indent_starts_special = ['for']
	indent_starts_special_good = ['then','else']
	indent_starts_ends = ['triggers']

	def create_indent(self):
		self.indent = []
		indent = 0
		for a in (self.commands):
			if a and not isinstance(a[0],list):
				if a[0] in self.indent_starts:
					indent += 1
			for i,x in enumerate(a):
				if x == 'end':
					indent -= 1
				elif x in self.indent_starts_anywhere:
					indent += 1
				elif x in self.indent_starts_special:
					if not i or (a[i-1] in self.indent_starts_special_good):
						indent += 1
				elif x in self.indent_starts_ends:
					if i == (len(a) - 1):
						indent += 1
			self.indent.append(indent)
		if indent:
			self.writer()
			self.error(indent)
			exit()
		return 1

	

	def create_commands(self):
		
		self.ar = []
		new_replaced = []
		self.replace_with_new = '0'
		counter = 0

		last_rep = 0
		do_rep = 0
		reps = []
		for a in self.txt.split(' '):
			last_rep = do_rep
			do_rep = 0
			
			if a == self.replace_with:
				to_add = self.replaced[counter]
				counter += 1
				do_rep = 1

			elif should_replace(a) or not is_kw(a):
				to_add = a
				do_rep = 1
				
			if do_rep:
				if last_rep:
					new_replaced[len(new_replaced) - 1].append(to_add)
				else:
					new_replaced.append([to_add])
					self.ar.append(self.replace_with)
				continue
			if a:
				self.ar.append(Alias(a.lower().strip()))
				
		self.replaced = new_replaced
		counter = 0
		self.commands = []
		for c in ' '.join(self.ar).split('.'):
			ar = c.strip().split(' ')
			if not ar: continue
			start_ar = []
			first = 0
			
			for i,x in enumerate(ar):
				if x == self.replace_with: 
					start_ar.append(counter)
					counter += 1
					first = i + 1
					continue
				
				break
			if first:
				self.commands.append(start_ar)
				ar = ar[first:]

				
			nc = []
			for a in ar:
				if a == self.replace_with:
					nc.append(counter)
					counter += 1
				else:
					nc.append(a)
			if nc:
				self.commands.append(nc)
		
		
	
		return 1
	


from jivi.oe.Imports import *
exec(Wrapper.main_setter)

from jivi.oe.ReaderParts.BigBlock import get as GetBigBlock



class Compiler(Wrapper):
	
	todofs = ['read_file_first_step','readit_first_step','writeit_first_step']
	todoss = ['read_file_ss','readit_ss','writeit_ss']
	
	todo = todofs
	def __init__(self,main,fp):
		set_main(main)

		self.fp = File.fp(fp).lower().strip()
		self.fp_compiled_first_step = SESSION.fp_compiled(self.fp,1)
		self.fp_compiled_ss = SESSION.fp_compiled(self.fp,0)
		
		self.name = self.fp + "\nCompiler"

		
	def read_file_ss(self):
		self.ar = File.jread(self.fp_compiled_first_step)
		
	def readit_ss(self):
		pass
		
	def writeit_ss(self):
		File.jwrite(self.fp_compiled_ss,self.ar)
	
	def read_file_first_step(self):
		self.txt = File.all_full(self.fp)
		if not self.txt: return 0

		
		return self.txt
		
		
		
		
		

				 
				
	def readit_first_step(self):
		self.ar = [[FILEPATH,self.fp]] + GetBigBlock(self.txt)
		

	def writeit_first_step(self):
		File.jwrite(self.fp_compiled_first_step,self.ar,indent=None)
		#File.jwrite(self.fp_compiled_first_step,make_readable_ar(self.ar))
		
		


from urllib.request import Request,urlopen
from time import sleep,time
from queue import Queue
from typing import List
from threading import Thread
from operator import itemgetter
import os,random, traceback, math, datetime, http
from jivi.util import to_string,to_int

class DownloadFileInternal:
	block_size            : int          = 0
	nthreads              : int          = 0
	url                   : str          = None
	fp                    : str          = None
	size                  : int          = 0
	accept_ranges         : bool         = False

	info_thread           : Thread     
	info_write_speed = 0
	info_download_speed = 0

	writer_bytes_written  : int          = 0
	writer_blocks_written : int          = 0
	writer_q              : Queue        = None
	writer_thread         : Thread      

	dblock_bytes_loaded   : int          = 0
	dblock_success        : int          = 0
	dblock_total          : int          = 0
	dblock_failed         : int          = 0
	dblock_q              : Queue        = None
	dblock_threads        : List[Thread] = []

	status_running        : bool         = False

        

	
	def __init__(self,url:str,fp:str,block_size_kb:int=500,nthreads:int=5):
		self.url             = url
		self.fp              = fp
		self.block_size      = block_size_kb * 1024
		self.nthreads        = nthreads
		
		self.info_init()
		self.dlbock_init()
		self.status_init()
		self.writer_init()

	def status_init(self):
		self.status_running  = False


	def dlbock_init(self):
		self.dblock_q        = Queue()
		self.dblock_threads  = [Thread(target=self.dblock_thread_func,daemon=True) for i in range(self.nthreads)]


	def dblock_reset(self):
		self.dblock_bytes_loaded = 0
		self.dblock_success      = 0
		self.dblock_total        = 0
		self.dblock_failed       = 0

	def dblock_start(self):
		self.dblock_reset()
		self.dblock_total = self.dblock_q.qsize()
		for t in self.dblock_threads:
			t.start()

	def dblock_thread_func(self):
		def download_block(index=0,failed=0):
			def action_success():
				self.dblock_success += 1
			def action_failed():
				if failed < 2:
					self.dblock_q.put(dict(index=index,failed=failed+1))
					return
				else:
					self.dblock_failed += 1
			
			block_start, block_stop = ((index*self.block_size), min( (self.size - 1) , ((index + 1) * self.block_size - 1) ) )

			data_len = block_stop - block_start + 1
			request    = None

			for attemp in range(5):
				try:
					request                  = Request(self.url)
					request.headers["Range"] = 'bytes=%d-%d'%(block_start,block_stop)
					break
				except:
					print(traceback.format_exc())
					sleep(3)
			
			if request is None:
				return action_failed()
			for attemp in range(10):
				try:
					data = urlopen(request)
					data = data.read()
					self.dblock_bytes_loaded += data_len
					self.writer_q.put(dict(index=index,data=data))
					return action_success()
				except (http.client.IncompleteRead) as e:
					data_len_to_do = (block_stop - block_start + 1)
					self.dblock_bytes_loaded += data_len_to_do
					data_to_write = e.partial + str.encode('0'*( data_len_to_do - len(e.partial)))
					self.writer_q.put(dict(index=index,data=data_to_write))
					
		
					return action_success()
				except:
					print(traceback.format_exc())
					sleep(2)
		
			return action_failed()
		
		while self.status_running and ( not self.status_finished ):
			while self.dblock_q.empty():
				if self.dblock_finished:
					return
				sleep(2)
			
			todo        = self.dblock_q.get()
			download_block(**todo)
	
	def writer_init(self):
		self.writer_q        = Queue()
		self.writer_thread   = Thread(target=self.writer_thread_func,daemon=True)

	def writer_reset(self):
		self.writer_bytes_written = 0
		self.writer_blocks_written = 0

	def writer_start(self):
		self.writer_reset() #!
		self.writer_thread.start()

	def writer_thread_func(self):
		def try_writing():
			todo = []
			while not self.writer_q.empty():
				q = self.writer_q.get()
				todo.append(q)
			
			if not todo:
				return
			todo.sort(key=lambda x: x.get('index'))

			with open(self.fp, "r+b") as f:
				cursor_now  = 0
				for td in todo:
					index,data 		   = itemgetter('index','data')(td)
					cursor_destination = index * self.block_size
					if (cursor_delta := cursor_destination-cursor_now) > 0:
						f.seek(cursor_delta,1)
					f.write(data)
					writtenNow    = len(data)
					cursor_now    = cursor_destination+writtenNow
					self.writer_bytes_written += writtenNow
					self.writer_blocks_written += 1

		
		while self.status_running and ( not self.status_finished ):
			try_writing()
			sleep(1)
			
	def info_init(self):
		self.info_thread     = Thread(target=self.info_thread_func,daemon=True)

	def info_start(self):
		self.info_thread.start()



	def info_thread_func(self):

		self.info_write_speed          = 0
		self.info_download_speed       = 0
		sleep_time = 1

		def get_data_now():
			return dict(
				t=time(),
				written=self.writer_bytes_written,
				downloaded=self.dblock_bytes_loaded
			)
		data = []
		data.append(get_data_now())
		sleep(1)
		while self.status_running and ( not self.status_finished ):
			data.append(get_data_now())
			data = data[-5:]

			first_data_entry = data[0]
			last_data_entry = data[-1]
			
			delta = {k : last_data_entry[k] - first_data_entry[k] for k in ['t','written','downloaded']}
			
			self.info_write_speed		= delta.get('written') / delta.get('t')
			self.info_download_speed	= delta.get('downloaded') / delta.get('t')

			sleep_time_more = max(0, sleep_time - (time() - last_data_entry.get('t')))
			if sleep_time_more:
				sleep(sleep_time_more)

	

	
	def reload(self):
		def reload_header():
			self.size          = 0
			self.accept_ranges = False
			for attempt in range(5):
				try:
					open_response = urlopen(self.url,timeout=10)
					self.size = to_int(open_response.headers.get('Content-Length',0))
					if not self.size:
						return False
					self.accept_ranges  = open_response.headers.get('accept-ranges','no') == 'bytes'
					return True
				except:
					sleep(1)
			
			return False
	
		def reload_local_file():
			def write_empty_file_if_not_exist():
				file_already_started 			   = os.path.isfile(self.fp) and (os.path.getsize(self.fp) == self.size)
				if file_already_started:
					return
				
				bytes_written 					   = 0
				bytes_to_write                     = self.size
				empty_str_20mb_len                 = 1024*1024*20
				empty_str_20mb                     = str.encode('0'*empty_str_20mb_len)
				with open(self.fp,"wb") as f:
					while bytes_written + empty_str_20mb_len <= bytes_to_write:
						f.write(empty_str_20mb)
						bytes_written += empty_str_20mb_len
					
					bytes_to_write_more = bytes_to_write -  bytes_written
					if bytes_to_write_more:

						f.write(str.encode('0'*bytes_to_write_more))
						

				sleep(1)

			def get_block_indexes_to_download():
				block_size             = self.block_size
				empty_block_str        = str.encode('0'*block_size)
				block_index            = 0
				indexes_to_download    = []
				blocks_already_written = 0
				
				with open(self.fp,"rb") as f:
					data = f.read(block_size)
					while data:
						data_len  = len(data)
						do_append = ((data_len == block_size) and (data == empty_block_str)) or ((data_len != block_size) and (data == str.encode('0'*data_len)))
						if do_append:
							indexes_to_download.append(block_index)
						else:
							blocks_already_written 	+= 1

						block_index += 1
						data 		= f.read(block_size)


				return indexes_to_download	

			def set_block_indexes_to_download(indexes):
				for i in indexes:
					self.dblock_q.put((dict(index=i,failed=0)))

			write_empty_file_if_not_exist()
			block_indexes_to_download = get_block_indexes_to_download()


			set_block_indexes_to_download(block_indexes_to_download)
		reload_header()
		reload_local_file()

	def stop(self):
		self.status_running = False
		return True
	
	def start(self):
		self.reload()
		self.status_running = True
		funcs_to_do = [self.dblock_start,self.writer_start,self.info_start]
		for t in funcs_to_do:
			t()
		

	def reset(self):
		print(f'Reset not yet implemented')

	def move(self,new_fp):
		print(f'Move not yet implemented!')
	
	@property
	def writer_finished(self) ->bool:
		return self.writer_blocks_written >= (self.dblock_total - self.dblock_failed)

	@property
	def dblock_finished(self) ->bool:
		return self.dblock_total == (self.dblock_success + self.dblock_failed)

	
	@property
	def status_finished(self) -> bool:
		return self.writer_finished and self.dblock_finished
	




from jivi.fs import *
from os import system
from time import time
from jivi.oe.constants import *
from jivi.oe.Reader import Reader

"""
	proc
	func
	vars
	mainblock
	
	widget
		...
		trigger
		
	



"""

zoek_ar = [(a,b,len(a)) for a,b in [('/*',1),('*/',-1),('"',2),("'",3),('{',5),('}',-5),('&glob',4),('&scop',4)]]
is_string = {3 : 1,2 : 1}
is_comment = {1 : 1, -1 : 1}
is_include = {5 : 1, -5 : 1}
is_scope = 4

replace_with = ' R'
replace_with_len = len(replace_with)
str_numbers = {str(i) : 1 for i in range(10)}
string_appenders = {':%s' % k : 1 for k in 'r,l,c,t,u'.split(',')}
class Filereader:
	functions = ['set_txt','strip_string_comment','write_all']
	
	def __init__(self,fp,env=0):
		self.errors = []
		self.base = File.base(fp)

		self.env = env

			
			

		
		
		self.fp = File.fp(fp)
		self.replaced = []
		self.txt = ""
		self.original = ""
		self.pre_defines = []
		for a in self.functions:
			if not getattr(self,a)():
				self.add_error('failed ' + a)
				break
		
	def set_write_fp(self):
		if self.env:
			pass
		else:
			self.fp_error = Dir.create("fre") + self.base
			self.fp_out = Dir.create("fro") + self.base
			
		File.delete(self.fp_error)
		File.delete(self.fp_out)
	def write_error(self):
		File.delete(self.fp_error)
		if not self.errors: return
		File.write(self.fp_error,"\n".join(self.errors))
		
	def write_all(self):
		self.set_write_fp()
		self.write_error()
		self.write_out()
		
	def write_out(self):
		
		File.write(self.fp_out,self.txt)
	
	def set_txt(self):
		self.txt = File.all(self.fp)
		if not self.txt: return False
		lines = []
		for line in self.txt.replace("\t"," ").splitlines():
			line = line.strip()
			linelow = line.lower()
			if linelow.startswith('&analyze'): continue
			lines.append(line)
			
		self.lines = []
		
		concat = 0
		for line in lines:
			old_concat = concat

			concat = line.endswith('~')
			if concat:
				line = line[:-1]
			
			if old_concat:
				self.lines[-1] += line
				continue
			
			self.lines.append(line)
			
		
				
				
				
		self.original = "\n".join(self.lines)
		self.txt = self.original[:]
		return 1

			
	def strip_string_comment(self):

			count = 0
			start_index = 0
			ar = []
			for needle,value,la in zoek_ar:
				index = self.txt.find(needle)
				while index != -1:
					ar.append((index,value))
					index = self.txt.find(needle,index+la)
			ar.sort()
			inum = 0
			lar = len(ar)
			larminone = lar - 1
			temp_ar = []
			last_stop = 0

			while inum < lar:
				start_inum = inum
				start_index,val = ar[inum]
				if val == is_scope:
					if (not start_index) or self.txt[start_index - 1:start_index] == "\n":
						stop = self.txt.find("\n",start_index)
						if stop == -1: stop = len(self.txt)
						self.pre_defines.append(self.txt[start_index:stop])
						start = start_index
						self.replaced.append((start,stop,0))
						if start != last_stop:
							temp_ar.append(self.txt[last_stop:start])
						last_stop = stop

					else:
						print(start_index)
						stop = self.txt.find("\n",start_index)
						if stop == -1: stop = len(self.txt)
						print(self.txt[start_index:stop])
						print(self.fp)
						exit()
						

				elif val in is_string:
					inum += 1
					string_val = val
					string_found = 0
					while inum < larminone:
						if (ar[inum][1] != string_val):
							inum += 1
						elif (ar[inum + 1][1] == string_val) and (ar[inum + 1][0] == (ar[inum][0] + 1)):
							inum += 2
						else:
							string_found = ar[inum][0]
							break
					if (not string_found) and (inum < lar) and (ar[inum][1] == string_val):
						string_found = ar[inum][0]
					if string_found and self.txt[string_found + 1:string_found + 3].lower() in string_appenders:
						string_found += 2
						while self.txt[string_found+1] in str_numbers:
							string_found += 1
					if not string_found:
						self.add_error_at_pos("do_str_c_pre : not string_found",start_index,func='add_all_strings')
						string_found = start_index
						inum = start_inum
					start = start_index
					stop = string_found + 1
					if start != last_stop:
						temp_ar.append(self.txt[last_stop:start])
					temp_ar.append(replace_with)
					self.replaced.append((start,stop,replace_with_len))
					last_stop = stop
				elif val in is_include:
					count = val
					inum += 1
					found = 0
					while inum < lar:
						index,val = ar[inum]
						if val in is_include:
							count += val
							if count < 0: print("Error !")
							if not count:
								start = start_index
								stop = index + 1
								if start != last_stop:
									temp_ar.append(self.txt[last_stop:start])
								temp_ar.append(replace_with)
								self.replaced.append((start,stop,replace_with_len))
								last_stop = stop
								found = 1
								break
						inum += 1
					if not found:
						self.add_error_at_pos("do_str_c_pre : is_include not found! started at line ",start_index)
				elif val in is_comment:
					last_index_plus_one = 0
					count = val
					inum += 1
					found_comment = 0
					while inum < lar:
						index,val = ar[inum]
						if val in is_comment and last_index_plus_one != index:
							count += val
							if count < 0: print("Error !")
							if not count:
								start = start_index
								stop = index + 2
								self.replaced.append((start,stop,0))
								if start != last_stop:
									temp_ar.append(self.txt[last_stop:start])
								last_stop = stop
								found_comment = 1
								break
							last_index_plus_one = index + 1
						inum += 1
					if not found_comment:
						self.add_error_at_pos("do_str_c_pre : is_comment not found_comment! started at line ",start_index)
						
				while inum < lar and (ar[inum][0] < last_stop):
					inum += 1
				
				
			temp_ar.append(self.txt[last_stop:])
			self.txt = "".join(temp_ar)
			return 1
			

	def get_line(self,pos):
		return self.original.count("\n",0,pos)

	def add_error_at_pos(self,s,pos,*a,**b):
		l = self.get_line(pos)
		if l != None:
			return self.add_error(s + "\nat line {l}".format(l=l+1),*a,**b)
		return self.add_error(s + "\n at pos " + str(pos),*a,**b)



	def add_error(self,s,*a,**b):
		self.errors.append(s)
		
		
	def make_replaced_two(self):
		self.replaced.sort()
		nr = []
		last_stop = -1
		last_start = -1
		for start,stop,l in self.replaced:
			if last_start < start < last_stop:
				last_stop = max(last_stop,stop)
				nr[-1] = (last_start,last_stop,1)
			else:
				nr.append((start,stop,l))
				last_stop = stop
				last_start = start
		self.replaced = nr

	def make_replaced(self):
		self.replaced.sort()
		nr = []
		last_stop = -1
		last_start = -1
		for start,stop,l in self.replaced:
			if last_start < start < last_stop:
				last_stop = max(last_stop,stop)
				nr[-1] = (last_start,last_stop,1)
			else:
				nr.append((start,stop,l))
				last_stop = stop
				last_start = start
		self.replaced = nr
		self.replaced_helper = []
		n = 0
		for start,stop,l in self.replaced:
			lengte = stop - start - l
			self.replaced_helper.append((start - n,n))
			n += (stop - start) - l
		self.replaced_helper.append((self.txt_o_len + 10,n))


SERIALIZE_NAME
BGCOLOR
COLUMN_LABEL
CONTEXT_HELP_ID



FONT

DCOLOR




FGCOLOR




FORMAT




DECIMALS

INITIAL







LABEL




MOUSE_POINTER

PFCOLOR

BEFORE_TABLE





















































from .src import Download
import os 
from time import sleep
from _thread import start_new_thread as Thread
from .. import util
from ..fs import JFile


URL = '''
http://praktijkdelotus.be/3.rar
'''.strip()

jFile = JFile(r'C:\temp\3.rar')
jFile.delete()





class DownloadTest(Download):
	def test(self):
		def data_test():
			to_print_ar = []

			prop_names = 'total,downloaded,written,to_download,to_write,speed'.split(',')
			for n in prop_names:
				to_print_ar.append([n,' : ',getattr(self.dataHuman,n)])

			for x in util.string.same_len_ar(to_print_ar):
				print(x)
			print('--------')

		def control_test():
			prop_names = 'start,stop,reset,move'.split(',')
			prop_names = 'stop,reset,move'.split(',')
			for n in prop_names:
				print(f'{n} : {getattr(self.control,n)()}')

			print('--------')

		data_test()
	





class DownloadTester(util.Tester):

	def test(self):
		def download_test_normal():
			downloader = DownloadTest(URL,jFile.fp)

			def printer():
				while True:
					downloader.test()
					sleep(5)

			downloader.control.start()

			Thread(printer,())

			while not downloader.status.finished:
				sleep(1)

		return [download_test_normal]
















from jivi.fs import *

class StripperGood:
	
	def __init__(self,env,fp):
		self.errors = []
		self.replaced = []
		self.commands = []
		self.scopes = []
		self.env = env
		self.fp = File.fp(fp)
		for do in self.dos:
			if not getattr(self,do)():
				self.error("failed to do {do}".format(do=do))
				break
		
		self.write()
		
	read_file_to_replace = ['&glob','&scop','&scop']
	

	def read_file(self):
		self.txt = File.all_full(self.fp)
		
		if not self.txt: return 0
		for a in self.read_file_to_replace:
			
			for i in range(1,len(a)+1):
				s = a[:i]
				up = s[:-1].lower() + s[-1].upper()
				self.txt = self.txt.replace(up,s.lower())
				
		
		return self.txt
		
		
	
	def set_replace_char(self):
		self.replace_with = 0
		self.replace_with_len = 0
		for i in range(255):
			x = chr(i).lower()
			if self.txt.find(x) == -1 and self.txt.find(x.upper()) == -1:
				self.replace_with = x
				self.replace_with_len = len(x)
				break
				
		return self.replace_with
	def error(self,s):
		print("Stripper\n{fp}\n{s}".format(fp=self.fp,s=s))

	def write(self):
		File.write("out\\" + File.base(self.fp),self.txt)


	def writer(self):
		File.jwrite('commands.txt',self.commands)
		File.jwrite('replaced.txt',self.replaced)
		self.write_indent()
	
	
			
	def get_line(self,pos):
		return self.txt_o.count("\n",0,pos)

	def add_error_at_pos(self,s,pos,*a,**b):
		l = self.get_line(pos)
		if l != None:
			return self.add_error(s + "\nat line {l}".format(l=l+1),*a,**b)
		return self.add_error(s + "\n at pos " + str(pos),*a,**b)

	def add_error(self,s,func=0):
		self.errors.append(s)

	def write_error(self,error_name,s,add=1):
		fp = "out\\" + self.name + "__" + error_name + self.ex
		with open(fp,"w") as f:
			f.write(s)
		if add:
			self.errors.append(fp)
		return fp

	def write_indent(self):
		to_write = []
		last_c = 0
		for i,c in enumerate(self.commands):
			cc = []
			for x in c:
				if isinstance(x,int):
					cc.append(" ".join(self.replaced[x]))
					last_c = x
				else:
					cc.append(x)
			
			to_write.append("\t"*self.indent[i] + " ".join(cc) + '.')
			if self.indent[i] < 0: break
			
		print(last_c)
		File.write('indent.w',"\n".join(to_write))


import random as _random



def fill_to_len(s:str,min_len:int,filler=' ',left_side=True):
	if len(s) >= min_len:
		return s
	if left_side:
		return filler*(min_len - len(s)) + s
	
	return s + filler*(min_len - len(s))

def between(s:str,a:str,b:str) -> str:
	try:
		x = s.find(a)
		y = s.find(b,x+len(a))
		retval = s[x+len(a):y]
		return retval
	except:
		return None


def random(n=10):
	letters = 'abcdefghijklmnopqrstuvwxyz0123456789'
	return ''.join(_random.choice(letters) for i in range(n))


def same_len_ar(ar,to_string=True):
	max_lens = []
	for i,a in enumerate(ar):
		if not i:
			max_lens = [len(a_el) for a_el in a]
		else:
			max_lens = [max(max_lens[a_ind],len(a[a_ind])) for a_ind in range(len(a))]
	
	ret_val = []
	for x in ar:
		ret_val.append(' '.join([fill_to_len(s,min_len=max_lens[index],left_side=False) for index,s in enumerate(x)]))
	

	return ret_val



from jivi import jv


class MenuTester(jv.util.Tester):
	def Menu(self):
		def test():
			menu = jv.menu.Menu(max_rows_auto=True)

			def maak_aantal_startend_met(aantal:int,startend:str):
				for i in range(aantal):
					naam = startend + jv.util.string.random(10)
					@menu.item(naam,exit_after=True)
					def print_test(p:jv.menu.Param):
						print(p.itemClicked)
						return True


			maak_aantal_startend_met(20,'jor ')
			maak_aantal_startend_met(20,'lot ')
			maak_aantal_startend_met(20,'jon ')
			maak_aantal_startend_met(20,'aka ')
			maak_aantal_startend_met(20,'deepthroat ')

			menu.display()

		return [test]




from Stripper import Stripper

from jivi.fs import *

#File.jwrite('kak.json',list(Dir.files(r'C:\develop\l_intex',ex='w',deep=1,full=1)))
# exit()
kak = File.jread('kak.json')
for i,fp in enumerate(kak):
	if i <= 247: continue	
	Stripper(0,fp)
	try:
		Stripper(0,fp)
	except:
		print(i)	
		print(fp)
		exit()
	print(i)



from .downloader.download_test import DownloadTester
from .fs.fs_test import FSTester
from .menu.menu_test import MenuTester
from .util.util_test import UtilTester
from .wrapper.wrapper_test import WrapperTester



class JiviTester:
	download = DownloadTester
	fs = FSTester
	menu = MenuTester
	util = UtilTester
	wrapper = WrapperTester

	@classmethod
	def reset(cls):
		cls.download = DownloadTester
		cls.fs = FSTester
		cls.menu = MenuTester
		cls.util = UtilTester
		cls.wrapper = WrapperTester



	@classmethod
	def only(cls,*names):
		ar = []
		for n in names:
			if isinstance(n,str):
				ar.append(getattr(cls,n,None))
			else:
				ar.append(n)

		return cls.do(ar)

	@classmethod
	def do(cls,ar=None):
		if ar is None:
			ar = [cls.download,cls.fs,cls.menu,cls.util,cls.wrapper]
		for x in ar:
			if not x:
				continue
			x.do()



from jivi.oe.constants import *
from jivi.fs import *

ALL_KW = {k : 1 for k in ALIAS.keys()}
for k in DEFINE_TYPES.keys(): ALL_KW[k] = 1


def is_kw(s):
	s = s.lower().strip()
	return s in ALL_KW
from time import time

strip_string_comments_preproc_ar = [(a,b,len(a)) for a,b in [('/*',1),('*/',-1),('"',2),("'",3),('{',5),('}',-5),('&glob',6),('&scop',7)]]
"""
eerst alle strings en comments weg daarna dubbele spaties
"""

is_number_val = 8


is_string = {3 : 1,2 : 1}
is_comment = {1 : 1, -1 : 1}
is_include = {5 : 1, -5 : 1}
is_scope = {6 : 1, 7 : 1}
string_appenders = {':%s' % k : 1 for k in 'r,l,c,t,u'.split(',')}
str_numbers = {str(i) : 1 for i in range(10)}
is_scope_local = 7


def is_decimal(x):
	try:
		s = float(x)
		return 1
	except:
		return 0
		
def should_replace(s):
	if s == '.': return 0
	if is_decimal('0' + s): return 1
	s = s.strip()
	ind = s.find('.',1) 
	if ind != -1 and ind != (len(s) - 1): return 1
	


class Stripper:
	dos = ['read_file','set_replace_char','strip_string_comments_preproc','strip_rest','create_commands_new']

	def __init__(self,env,fp):
		self.errors = []
		self.replaced = []
		self.commands = []
		self.scopes = []
		self.env = env
		self.fp = File.fp(fp)
		for do in self.dos:
			if not getattr(self,do)():
				self.error("failed to do {do}".format(do=do))
				break
		
		self.write()
		
	read_file_to_replace = ['&glob','&scop']
	

	
	def read_file(self):
		self.txt = File.all_full(self.fp)
		
		if not self.txt: return 0
		for a in self.read_file_to_replace:
			
			for i in range(1,len(a)+1):
				s = a[:i]
				up = s[:-1].lower() + s[-1].upper()
				self.txt = self.txt.replace(up,s.lower())
				
		
		return self.txt
		
		
	
	def set_replace_char(self):
		self.replace_with = 0
		self.replace_with_len = 0
		for i in range(255):
			x = chr(i).lower()
			if self.txt.find(x) == -1 and self.txt.find(x.upper()) == -1:
				self.replace_with = x
				self.replace_with_len = len(x)
				break
				
		return self.replace_with
	def error(self,s):
		print("Stripper\n{fp}\n{s}".format(fp=self.fp,s=s))

	def write(self):
		File.write("out\\" + File.base(self.fp),self.txt)

		
	def strip_string_comments_preproc(self):
		ignore_indexes = {}
		index = self.txt.find('~')
		while index != -1:
			ignore_indexes[index+1] = 1
			index = self.txt.find('~',index+1)
		
		
		
		count = 0

		ar = []
		for needle,value,la in strip_string_comments_preproc_ar:
			index = self.txt.find(needle)
			while index != -1:
				ar.append((index,value))
				index = self.txt.find(needle,index+la)
		
		
		ar.sort()
		inum = 0
		lar = len(ar)
		larminone = lar - 1
		temp_ar = []
		
		add_to_replaced = 0
		start_index_replaced = 0
		stop_index_replaced = 0
		last_stop = 0
		while inum < len(ar):
			
			start_inum = inum
			start_index_replaced,val = ar[inum]
			stop_index_replaced = 0
			if val in is_scope:
				stop_index_replaced = self.txt.find("\n",start_index_replaced)
				shouldbreak = 0
				if stop_index_replaced == -1: 
					stop_index_replaced = len(self.txt)
					shouldbreak = 1
				
				r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
					
						
				while r.endswith('~'):
					r = r[:-1]
					if shouldbreak:
						break
					stop_index_replaced = self.txt.find("\n",stop_index_replaced + 1)
					shouldbreak = 0
					if stop_index_replaced == -1: 
						stop_index_replaced = len(self.txt)
						shouldbreak = 1
					r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
				
				add_to_replaced = 0
				self.scopes.append(((val == is_scope_local),r))
		
				
			elif val in is_string:
				add_to_replaced = 1
				inum += 1
				string_val = val
				string_found = 0
				while inum < larminone:
					if (ar[inum][1] != string_val) or (ar[inum][0] in ignore_indexes):
						inum += 1
					elif (ar[inum + 1][1] == string_val) and (ar[inum + 1][0] == (ar[inum][0] + 1)):
						inum += 2
					else:
						string_found = ar[inum][0]
						break
				if (not string_found) and (inum < lar) and (ar[inum][1] == string_val):
					string_found = ar[inum][0]
				if string_found and self.txt[string_found + 1:string_found + 3].lower() in string_appenders:
					string_found += 2
					
					while self.txt[string_found+1] in str_numbers:
						string_found += 1
				if not string_found:
					self.add_error_at_pos("do_str_c_pre : not string_found",start_index_replaced,func='add_all_strings')
					string_found = start_index_replaced
					inum = start_inum
				
				start = start_index_replaced
				stop_index_replaced = string_found + 1

			elif val in is_include:
				add_to_replaced = 1
				count = val
				inum += 1
				found = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_include:
						count += val
						if count < 0: print("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 1

							found = 1
							break
					inum += 1
				if not found:
					self.add_error_at_pos("do_str_c_pre : is_include not found! started at line ",start_index_replaced)
			elif val in is_comment:
				add_to_replaced = 0
				last_index_plus_one = 0
				count = val
				inum += 1
				found_comment = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_comment and last_index_plus_one != index:
						count += val
						if count < 0: print("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 2
							found_comment = 1
							break
						last_index_plus_one = index + 1
					inum += 1
				if not found_comment:
					self.add_error_at_pos("do_str_c_pre : is_comment not found_comment! started at line ",start_index_replaced)
					
					
			while inum < lar and (ar[inum][0] < stop_index_replaced):
				inum += 1
			if last_stop != start_index_replaced:
				temp_ar.append(self.txt[last_stop:start_index_replaced])
			if add_to_replaced:
				self.replaced.append(self.txt[start_index_replaced:stop_index_replaced])
				temp_ar.append(self.replace_with)
			last_stop = stop_index_replaced
				
		temp_ar.append(self.txt[stop_index_replaced:])
		self.txt = "".join(temp_ar)
		return len(self.errors) == 0
		
		
	strip_rest_replacers = ['[',']',',','. ']
	def strip_rest(self):
		index_now = self.txt.find('&')
		while index_now != -1:

			stop_index = self.txt.find("\n",index_now)
			shouldbreak = 0
			if stop_index == -1: 
				stop_index = len(self.txt)
				shouldbreak = 1
			
			r = self.txt[index_now:stop_index].rstrip()
				
					
			while r.endswith('~'):
				r = r[:-1]
				if shouldbreak:
					break
				stop_index = self.txt.find("\n",stop_index + 1)
				shouldbreak = 0
				if stop_index == -1: 
					stop_index = len(self.txt)
					shouldbreak = 1
				r = self.txt[index_now:stop_index].rstrip()
				
			self.txt = self.txt[:index_now] + self.txt[stop_index:]
			index_now = self.txt.find('&')
		self.txt = self.txt.replace("\t"," ").replace("\n"," ").lower().replace(self.replace_with,' ' + self.replace_with + ' ').replace(':','.')
		for a in self.strip_rest_replacers:
			self.txt = self.txt.replace(a,' ' + a + ' ')
		while self.txt.find('  ') != -1:
			self.txt = self.txt.replace('  ',' ')
		self.txt = self.txt.strip()
		return 1
	def create_commands_new(self):
		
		self.ar = []
		new_replaced = []
		self.replace_with_new = '0'
		counter = 0

		last_rep = 0
		do_rep = 0
		for a in self.txt.split(' '):
			last_rep = do_rep
			do_rep = 0
			
			if a == self.replace_with:
				to_add = self.replaced[counter]
				counter += 1
				do_rep = 1

			elif should_replace(a) or not is_kw(a):
				to_add = a
				do_rep = 1
				
			if do_rep:
				if last_rep:
					new_replaced[-1] += ' ' + to_add
				else:
					new_replaced.append(to_add)
					self.ar.append(self.replace_with)
				continue
			if a:
				self.ar.append(a)
				
		self.replaced = new_replaced
		counter
		self.commands = []
		for c in ' '.join(self.ar).split('.'):
			nc = []
			for a in c.strip().split(' '):
				if a == self.replace_with:
					nc.append(counter)
					counter += 1
				else:
					nc.append(a)
			if nc:
				self.commands.append(nc)
		
		return 1
		
	def create_commands_new_old(self):
		
		self.ar = []
		new_replaced = []
		self.replace_with_new = '0'
		counter = 0
		for a in self.txt.split(' '):
			if a == self.replace_with:
				new_replaced.append(self.replaced[counter])
				counter += 1
				self.ar.append(self.replace_with_new)
				continue
			if should_replace(a) or not is_kw(a):
				new_replaced.append(a)
				self.ar.append(self.replace_with_new)
				continue
			if a:
				self.ar.append(a)
				
		self.replaced = new_replaced
		
		self.commands = [c.strip().split(' ') for c in ' '.join(self.ar).split('.')]
		File.jwrite('commands.txt',self.commands)
		
				
	def create_commands(self):

		self.ar = []
		new_replaced = []
		
		counter = 0
		for a in self.txt.split(' '):
			if a == self.replace_with:
				new_replaced.append(self.replaced[counter])
				counter += 1
				self.ar.append(self.replace_with)
				continue
			if should_replace(a):
				new_replaced.append(a)
				self.ar.append(self.replace_with)
				continue
			if a:
				self.ar.append(a)
				
		self.replaced = new_replaced
		
		self.commands = [c.strip().split(' ') for c in ' '.join(self.ar).split('.')]
		File.jwrite('commands.txt',self.commands)
		File.jwrite('replaced.txt',self.replaced)
	
	def get_line(self,pos):
		return self.txt_o.count("\n",0,pos)

	def add_error_at_pos(self,s,pos,*a,**b):
		l = self.get_line(pos)
		if l != None:
			return self.add_error(s + "\nat line {l}".format(l=l+1),*a,**b)
		return self.add_error(s + "\n at pos " + str(pos),*a,**b)

	def add_error(self,s,func=0):
		self.errors.append(s)

	def write_error(self,error_name,s,add=1):
		fp = "out\\" + self.name + "__" + error_name + self.ex
		with open(fp,"w") as f:
			f.write(s)
		if add:
			self.errors.append(fp)
		return fp




from jivi.fs import *
from jivi.oe.Imports import *
"""
	strip comments
	split into
		string
		includes
		pp define

NONE                                               = 0
UNKNOWN                                            = 1
KEYWORD                                            = 2
STRING                                             = 3
STRING_CONCAT                                      = 4
NUMBER                                             = 5
TABLE                                              = 6
FIELD                                              = 7
VAR                                                = 8
BOOLEAN                                            = 9
EXPRESSION                                         = 10
PP_UNKNOWN                                         = 11
PP_SCOP                                            = 12
PP_GLOB                                            = 13
PP_UNDEF                                           = 14
PP_VAR                                             = 15
PP_VAR_CONCAT                                      = 16
PP_PAR                                             = 17
PP_INCL                                            = 18
PP_CONDITION                                       = 19
PP_CONDITIONBLOCK                                  = 20
COMMAND                                            = 21
FILEPATH                                           = 22


"""
from jivi.oe.Imports import *
get_scoped_string_re = re.compile("({([^}{]+)})")
is_scope_local = 7
is_scope_global = 6
is_undefine = 9
is_string = {3 : 1,2 : 1}
is_comment = {1 : 1, -1 : 1}
is_include = {5 : 1, -5 : 1}
is_scope = {is_scope_global : 1, is_scope_local : 1}
preproc_if = 11
preproc_elseif = 21
preproc_else = 31
preproc_endif = 41
preproc_then = 51
is_ppc = {k : 1 for k in [preproc_else,preproc_elseif,preproc_endif,preproc_if,preproc_then]}
string_comment_scopes_ar_ppc = [("/*",1,2),("*/",-1,2),("\"",2,1),("'",3,1),("{",5,1),("}",-5,1),("&glob",6,5),("&scop",7,5),("&undef",9,6),("&if",11,3),("&elseif",21,7),("&else",31,5),("&then",51,5),("&endif",41,6)]
string_appenders = {':%s' % k : 1 for k in 'r,l,c,t,u'.split(',')}
str_numbers = {str(i) : 1 for i in range(10)}
pp_if = "&if"
pp_elseif = "&elseif"
pp_else = "&else"
pp_endif = "&endif"
pp_then = "&then"

def get_find_ar(txt,needlevals):
	ar = []
	for needle,value in needlevals.items():
		la = len(needle)
		index = txt.find(needle)
		while index != -1:
			ar.append((index,value))
			index = txt.find(needle,index+la)

		
	return sorted(ar)
		
		
def create_pp_var_concat(ar):

	new_ar = []
	ret_ar = [PP_VAR_CONCAT,new_ar]
	s = ar[0]

	find_ar = get_find_ar(s,{"{" : 1, "}" : -1})
	l = len(find_ar)
	i = 0
	count = 0
	last_stop_index = 0
	while i < l:
		index,value = find_ar[i]
		stop_index = 0
		i += 1
		if value == 1:
			count = 1
			start_index = index
			while i < l:
				index,value = find_ar[i]
				count += value
				i += 1
				if not count: 
					stop_index = index + 1
					break
			
			if start_index != last_stop_index:
				new_ar.append([STRING,[s[last_stop_index:start_index],NONE]])
			last_stop_index = stop_index
				
			new_ar.append([PP_VAR,s[start_index:stop_index][2:-1].lower().strip()])
			
	s = s[last_stop_index:]
	if s:
		new_ar.append([STRING,[s,NONE]])
		

	do_strings(new_ar)

	return ret_ar
		
	
def create_string_concat(ar):
	new_ar = []
	ret_ar = [STRING_CONCAT,new_ar]
	s = ar[0]
	
	find_ar = get_find_ar(s,{"{" : 1, "}" : -1})
	l = len(find_ar)
	i = 0
	count = 0
	last_stop_index = 0
	while i < l:
		index,value = find_ar[i]
		stop_index = 0
		i += 1
		if value == 1:
			count = 1
			start_index = index
			while i < l:
				index,value = find_ar[i]
				count += value
				i += 1
				if not count: 
					stop_index = index + 1
					break
			
			if start_index != last_stop_index:
				new_ar.append([STRING,[s[last_stop_index:start_index],NONE]])
			last_stop_index = stop_index
				
			new_ar.append([PP_VAR,s[start_index:stop_index][2:-1].lower().strip()])
			
	s = s[last_stop_index:]
	if s:
		new_ar.append([STRING,[s,NONE]])
		


	do_strings(new_ar)

	return ret_ar
	


	
def do_strings(ar):
	for i,a in enumerate(ar):
		if not isinstance(a,list): continue
		if a[0] == STRING:
			index = a[1][0].find("{&")
			if index == -1: continue
			ar[i] = create_string_concat(a[1])
		elif a[0] == PP_VAR:
			index = a[1].find("{&")
			if index == -1: continue
			ar[i] = create_pp_var_concat([a[1]])
		elif a[0] == PP_CONDITIONBLOCK:
			for j,b in enumerate(a[1][0]):
				do_strings(ar[i][1][0][j])
			for j,b in enumerate(a[1][1]):
				do_strings(ar[i][1][1][j])
			
				
			
class PP_Conditionblock:
	condition_starts = {k : 1 for k in [pp_if,pp_elseif,pp_else]}
	condition_stops = {k : 1 for k in [pp_else,pp_elseif]}

	@staticmethod
	def has_ppc(ar):
		for i,a in enumerate(ar):
			if isinstance(a,list) and a[0] == PP_CONDITION:
				return i + 1

	@staticmethod
	def find_stop(ar,start=0):
		started = 0
		count = 0
		for i in range(start,len(ar)):
			x = ar[i]
			if isinstance(x,list) and x[0] == PP_CONDITION:
				if x[1] == "&if":
					count +=1
				elif x[1] == "&endif":
					count -= 1
				if count and (not started): started = 1
				if started and (not count):
					return i + 1

	def __init__(self,ar):
		self.ori = ar[:]
		self.ar = ar[:-1]
		self.condition_statements = []
		self.code_blocks = []
		self.ret_ar = [[PP_CONDITIONBLOCK,[self.condition_statements,self.code_blocks]]]

	def do(self):
		while self.ar:
			ll = len(self.ar)
			self.read_condition()
			self.read_code_block()
			if ll == len(self.ar):
				print("Failing ! PP_Conditionblock")
				print_ar(self.ar)
				print_ar(self.ori)
				exit()
		for i,block in enumerate(self.code_blocks):
			start_index = PP_Conditionblock.has_ppc(block)
			while start_index:
				start_index -= 1
				stop_index = PP_Conditionblock.find_stop(block,start=start_index)
				if not stop_index:
					print("do_ppc stop_index not found ! \n" + str(block[start_index:]))
					exit()
				self.code_blocks[i] = block[:start_index] + PP_Conditionblock(block[start_index:stop_index]).do() + block[stop_index:]
				start_index = PP_Conditionblock.has_ppc(self.code_blocks[i])
		return self.ret_ar

	def read_code_block(self):
		count = 0
		for i,a in enumerate(self.ar):
			if not (isinstance(self.ar[i],list) and self.ar[i][0] == PP_CONDITION): continue
			x = self.ar[i]
			if x[1] == pp_if:
				count += 1
				continue
			if x[1] in self.condition_stops:
				if not count:
					self.code_blocks.append(self.ar[:i])
					self.ar = self.ar[i:]
					return
				continue
			if x[1] == pp_endif:
				count -= 1
				continue
		if not count:
			self.code_blocks.append(self.ar)
			self.ar = []

	def read_condition(self):
		if not (self.ar and isinstance(self.ar,list) and self.ar[0][0] == PP_CONDITION and self.ar[0][1] in self.condition_starts): return
		if self.ar[0][1] == pp_else:
			self.ar = self.ar[1:]
			return
		count = 0
		for i,a in enumerate(self.ar):
			if not i: continue
			if not (isinstance(self.ar[i],list) and self.ar[i][0] == PP_CONDITION): continue
			x = self.ar[i]
			if x[1] == pp_if:
				count += 1
				continue
			if x[1] == pp_then:
				if not count:
					cs = self.ar[1:i]
					self.ar = self.ar[i+1:]
					#print(make_readable_ar(self.ar))
					#print(make_readable_ar(cs))
					self.condition_statements.append(cs)
					return
				continue
			if x[1] == pp_endif:
				count -= 1
				continue
		return None

class Reader:
	def __init__(self,txt):
		self.txt = txt
		self.read()

	def find_all_ar(self,x):
		ar = []
		index = self.txt.find(x)
		while index != -1:
			ar.append(index)
			index = self.txt.find(x,index+1)
		return ar
	read_file_to_replace = ['&glob','&scop','&undef','&elseif','&then','&endif','&if']

	def read(self):
		if not self.txt: return 0
		for a in self.read_file_to_replace:
			for i in range(1,len(a)+1):
				s = a[:i]
				up = s[:-1].lower() + s[-1].upper()
				self.txt = self.txt.replace(up,s.lower())
		return self.txt

	def scs_get_ar(self):
		ar = []
		for needle,value,la in string_comment_scopes_ar_ppc:
			index = self.txt.find(needle)
			while index != -1:
				ar.append((index,value))
				index = self.txt.find(needle,index+la)
		
		return sorted(ar)
	strip_rest_replacers = ['[',']',',','. ',' .','(',')','=']
	strip_rest_replacers_ends = 'procedure,function,case,for,triggers,finally'.split(',')

	def parse_include(self,txt):
		para_tab = {}
		para_ar = []
		ar = get(txt)
		if not ar: return []
		fp = ar[0]
		if isinstance(fp,list):
			if fp[0] == UNKNOWN:
				fp = fp[1]
			elif fp[0] == STRING:
				fp = fp[1][0]
			else:
				self.error("weird parse_include " + str(fp))
				exit()
		else:
			self.error("weird parse_include " + str(fp))
			exit()
		ar = ar[1:]
		if ar:
			ar = flat_ar(ar)
			while ar:
				if ar[0].startswith("&"):
					if len(ar) > 2 and ar[1] == "=":
						para_tab[ar[0][1:].lower().strip()] = ar[2]
						para_ar.append(ar[2])
						ar = ar[3:]
						continue
					else:
						self.error("weird parse_include " + str(txt))
						exit()
				elif ar[0] == "=":
					self.error("weird parse_include === " + str(txt))
					exit()
				else:
					para_ar.append(ar[0])
					ar = ar[1:]
		full_tab = {}
		for k,v in para_tab.items():
			full_tab["&" + k] = v
		for i,a in enumerate(para_ar):
			full_tab[str(i + 1)] = a
		return [[PP_INCL,[fp,full_tab]]]

	def parse_unknown(self,s):
		sar = [""]
		for l in s.split("\n"):
			if l.lower().strip().startswith("&ana"): continue
			sar.append(l)
		sar.append("")
		txt = " ".join(sar).lower().replace("\t"," ").replace(': ','. ')
		for a in self.strip_rest_replacers:
			txt = txt.replace(a,' ' + a + ' ')
		while txt.find('  ') != -1:
			txt = txt.replace('  ',' ')
		for a in self.strip_rest_replacers_ends:
			txt = txt.replace(' end ' + a,' end ')
		txt = txt.replace(' forward ',' end ')
		while txt.find('  ') != -1:
			txt = txt.replace('  ',' ')
		while txt.find('..') != -1:
			txt = txt.replace('..','.')
		ar = []
		for a in txt.strip().split(" "):
			a = a.lower().strip()
			if not a: continue
			kw = Keyword(a)
			if kw != None:
				ar.append(kw)
			else:
				ar.append([UNKNOWN,a])
		return ar

	def get_it(self):
		self.file_scopes = {}
		ignore_indexes = {}
		for a in self.find_all_ar('~'):
			if not a in ignore_indexes:
				ignore_indexes[a+1] = 1
		ar = self.scs_get_ar()
		count = 0
		inum = 0
		lar = len(ar)
		larminone = lar - 1
		self.new_txt_ar = []
		to_add = 0
		start_index_replaced = 0
		stop_index_replaced = 0
		last_stop = 0

		def get_next_stop_blank_or_new_line(index):
			rindex = len(self.txt)
			for a in [" ","\n","."]:
				ind = self.txt.find(a,index)
				if ind != -1 and ind < rindex: rindex = ind
			return rindex

		def read_until_blank(index):
			while self.txt[index+1:index+2] == " ":
				index += 1
			return index
		while inum < lar:
			start_inum = inum
			start_index_replaced,val = ar[inum]
			stop_index_replaced = 0
			if val == is_undefine:
				stop_index_replaced = self.txt.find(" ",start_index_replaced)
				stop_index_replaced = read_until_blank(stop_index_replaced)
				found_index = stop_index_replaced
				stop_index_replaced = get_next_stop_blank_or_new_line(found_index + 1)
				to_add = [[PP_UNDEF,self.txt[found_index+1:stop_index_replaced].lower().strip()]]
			elif val in is_ppc:
				stop_index_replaced = get_next_stop_blank_or_new_line(start_index_replaced)
				txt = self.txt[start_index_replaced:stop_index_replaced].lower().strip()
				to_add = [[PP_CONDITION,txt]]
			elif val in is_scope:
				stop_index_replaced = self.txt.find("\n",start_index_replaced)
				shouldbreak = 0
				if stop_index_replaced == -1:
					stop_index_replaced = len(self.txt)
					shouldbreak = 1
				r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
				while r.endswith('~'):
					if shouldbreak:
						break
					stop_index_replaced = self.txt.find("\n",stop_index_replaced + 1)
					shouldbreak = 0
					if stop_index_replaced == -1:
						stop_index_replaced = len(self.txt)
						shouldbreak = 1
					r = self.txt[start_index_replaced:stop_index_replaced].rstrip()
				scopename = 0
				scopenamear = r.replace('~','').split(' ')
				scopevalue = 0
				for i,x in enumerate(scopenamear):
					if not i: continue
					scopename = x.lower().strip()
					if not scopename: continue
					scopevalue = ' '.join(scopenamear[i+1:])
					break
				type = PP_GLOB
				if val == is_scope_local:
					type = PP_SCOP
				to_add = [[type,[scopename,scopevalue]]]
			elif val in is_string:
				inum += 1
				string_val = val
				string_found = 0
				string_txt = 0
				string_format = NONE
				while inum < larminone:
					if (ar[inum][0] in ignore_indexes):
						next_ind = ar[inum][0] + 1
						inum += 1
					elif (ar[inum][1] != string_val):
						inum += 1
					elif (ar[inum + 1][1] == string_val) and (ar[inum + 1][0] == (ar[inum][0] + 1)):
						inum += 2
					else:
						string_found = ar[inum][0]
						break
				if (not string_found) and (inum < lar) and (ar[inum][1] == string_val):
					string_found = ar[inum][0]
				if string_found:
					string_txt = self.txt[start_index_replaced:string_found+1]
				if string_found and self.txt[string_found + 1:string_found + 3].lower() in string_appenders:
					string_format = self.txt[string_found + 2:string_found + 3].lower()
					string_found += 2
					while self.txt[string_found+1] in str_numbers:
						string_found += 1
						string_format += self.txt[string_found+1]
				if not string_found:
					x = self.txt.find("\n",start_index_replaced + 1)
					self.info(self.txt[start_index_replaced:x])
					self.add_error_at_pos("do_str_c_pre : not string_found",start_index_replaced)
					string_found = start_index_replaced
					string_txt = self.txt[start_index_replaced:string_found+1]
					inum = start_inum
				start = start_index_replaced
				stop_index_replaced = string_found + 1
				string_txt = string_txt[1:-1]
				to_add = [[STRING,[string_txt,string_format]]]
			elif val in is_include:
				count = val
				inum += 1
				found = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_include:
						count += val
						if count < 0: self.error("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 1
							found = 1
							break
					inum += 1
				if not found:
					self.add_error_at_pos("do_str_c_pre : is_include not found! started at line ",start_index_replaced)
				txt = self.txt[start_index_replaced:stop_index_replaced][1:-1].strip()
				if txt[:1] == '&':
					to_add = [[PP_VAR,txt[1:].lower()]]
				else:
					to_add = self.parse_include(txt)
			elif val in is_comment:
				to_add = []
				last_index_plus_one = 0
				count = val
				inum += 1
				found_comment = 0
				while inum < lar:
					index,val = ar[inum]
					if val in is_comment and last_index_plus_one != index:
						count += val
						if count < 0: self.error("Error !")
						if not count:
							start = start_index_replaced
							stop_index_replaced = index + 2
							found_comment = 1
							break
						last_index_plus_one = index + 1
					inum += 1
				if not found_comment:
					x = self.txt.find("\n",start_index_replaced + 1)
					self.error(self.txt[start_index_replaced:x])
					self.add_error_at_pos("do_str_c_pre : is_comment not found_comment! started at line ",start_index_replaced)
			else:
				 self.error('val not found')
				 self.error(val)
			while (start_inum == inum) or (inum < lar and (ar[inum][0] < stop_index_replaced)):
				inum += 1
			rest_txt = self.txt[last_stop:start_index_replaced].strip()
			if rest_txt:
				self.new_txt_ar += self.parse_unknown(rest_txt)
			if to_add:
				self.new_txt_ar += to_add
			last_stop = stop_index_replaced
		rest_txt = self.txt[stop_index_replaced:].strip()
		if rest_txt:
			self.new_txt_ar += self.parse_unknown(rest_txt)
		while self.has_ppc():
			self.do_ppc()
		do_strings(self.new_txt_ar)
		return self.new_txt_ar

	
		

	def do_ppc(self):
		count = 0
		ppc_stop_index = PP_Conditionblock.find_stop(self.new_txt_ar,start=self.ppc_index)
		if not ppc_stop_index:
			self.error("do_ppc ppc_stop_index not found ! \n" + str(self.new_txt_ar[self.ppc_index:]))
			exit()
		self.new_txt_ar = self.new_txt_ar[:self.ppc_index] + PP_Conditionblock(self.new_txt_ar[self.ppc_index:ppc_stop_index]).do() + self.new_txt_ar[ppc_stop_index:]

	def has_ppc(self):
		for i,a in enumerate(self.new_txt_ar):
			if isinstance(a,list) and a[0] == PP_CONDITION:
				self.ppc_index = i
				return 1

def get(txt,nice=0):
	ar = Reader(txt).get_it()
	if nice:
		return make_readable_ar(ar)
	return ar


import jivi.util as jvUtil
from jivi.wrapper import jvWrapper,jvWrapperSuper
import jivi.menu as jvMenu
import jivi.fs as jvFs



class jv:
	menu = jvMenu
	wrapper = jvWrapper
	util = jvUtil
	fs = jvFs
	




class Object:
	
	def __init__(self,env,type,name):
		self.env = env
		self.type = type
		self.name = name
		
		
		
	def html(self):
		return "no html"
		
		
	def error(self,s):
		print("[{type}]{name} : {s}".format(type=self.type,name=self.name,s=s))
	
	


from jivi.pr.Object import Object

from jivi.pr.objecth.Variable import Variable


	



			
	
class Field(Variable):


	def __init__(self,*a,**b):
		table = 0
		if 'table' in b:
			table = b['table']
			del b['table']
		
		Variable.__init__(self,*a,**b)
		if table:
			self.table = table
			self.table.register_field(self)
			

class Table(Object):
	attr = []
	attr.append(dict(n='name',r=1,a='_file-name',lower=1))

	def __init__(self,*a,**b):
		self.fields = {}
		Object.__init__(self,*a,**b)
		
		
	
	def register_field(self,f):
		self.fields[f.name] = f


import math
import os
import subprocess
import traceback
import json
import time
import shutil
from _thread import start_new_thread as Thread
perr = traceback.print_exc

class JiviCore:
	DEFAULT_ENCODING = "cp850"

	@classmethod
	def cmd(cls,cmd:str):
		process = subprocess.Popen(cmd, shell = True)
		stdout, stderr = process.communicate()
	
	@classmethod
	def decode(cls,b : bytes,encoding=None) -> str:
		encoding_to_try = ([encoding,cls.DEFAULT_ENCODING] if encoding is not None else [cls.DEFAULT_ENCODING]) + ['cp1252','ascii','utf-8','latin-1']
		for a in encoding_to_try:
			try:
				s = b.decode(a)
				return s
			except:
				pass

	@classmethod
	def popen(cls,s:str,as_gen=True,stripped=True) -> [str]:
		def func():
			try:
				process 		= subprocess.Popen(s, stdout=subprocess.PIPE,shell=1)
				stdout, stderr 	= process.communicate()
				if stderr: 
					print(stderr)
				
				stdout = cls.decode(stdout)
				for x in stdout.splitlines():
					yield x.strip() if stripped else x
			except:
				pass
		if not as_gen:
			return list(func())
		return func()

class JiviDir:
	@classmethod
	def fp(cls,*inp,d=None,with_sep=True):
		
		fp_parts 	= [cls.fp_sepper(x,True) for x in (inp if d is None else [d] + list(inp))]
		fp 			= os.path.join(*fp_parts)
		return cls.fp_sepper(os.path.abspath(fp),with_sep=with_sep)
	
	@classmethod
	def fp_sepper(cls,fp,with_sep=True):
		return fp.rstrip(os.sep) + os.sep if with_sep else fp.rstrip(os.sep)
	
	@classmethod
	def cwd(cls,with_sep=True):
		return cls.fp(os.getcwd(),with_sep=with_sep)
	
	@classmethod
	def ends_with_sep(cls,fp:str):
		return fp.endswith(os.sep)
	
	@classmethod
	def up(cls,fp,n=1):
		return cls.fp(fp + os.sep,('..' + os.sep)*n,with_sep=cls.ends_with_sep(fp))

	@classmethod
	def name(cls,fp):
		return os.path.basename(cls.fp(fp).rstrip(os.sep))
	
	@classmethod
	def rel(cls,fp,b,with_sep=True):
		return cls.fp_sepper(os.path.relpath(fp,b),with_sep)
	
	@classmethod
	def find_file(cls,fp,fn,deep=1):
		flags = ['b','a-d']
		if deep: 
			flags.append('s')
		
		fp    = cls.fp(fp,with_sep=True)
		flags = JiviAr.join_start_too(flags,'/')
		cmd   = f'dir {fp}{fn} {flags} 2>nul'

		return JiviCore.popen(cmd,as_gen=True)

	@classmethod
	def find_dir(cls,fp,name,deep=1):
		fp    = cls.fp(fp,with_sep=True)

		if not deep:
			dp_to_find = cls.fp(fp,name)
			if cls.exists(dp_to_find):
				return [dp_to_find]
			return []
		
		
		flags = JiviAr.join_start_too(['b','ad','s'],'/')

		cmd   = f'dir {fp}{name} {flags} 2>nul'

		return JiviCore.popen(cmd,as_gen=True)
	
	@classmethod
	def files(cls,fp,full=1,ex=['*'],deep=None,img=None):
		if img:
			ex = ['bmp','jpg','jpeg','gif','png']
		fp    = cls.fp(fp,with_sep=True)
		flags = ['b','a-d']

		if deep: 
			flags.append('s')
			full = 0
	

		if isinstance(ex,str):
			ex = [ex]
		
		x 	  = " ".join([f'"{fp}*.{e}"' for e in ex])
		flags = JiviAr.join_start_too(flags,'/')
		cmd   = f'dir {x} {flags} 2>nul'

		for x in JiviCore.popen(cmd,as_gen=True):
			yield fp + x if full else x
		
	@classmethod
	def dirs(cls,fp,full=1,deep=None):
		fp    = cls.fp(fp,with_sep=True)
		flags = ['b','ad']
		if deep: 
			flags.append('s')
			full = 0
	
		flags = JiviAr.join_start_too(flags,'/')
		cmd   = f'dir {fp} {flags} 2>nul'

		for x in JiviCore.popen(cmd,as_gen=True):
			yield fp + x if full else x

	@classmethod
	def exists(cls,fp):
		return os.path.isdir(fp)


	@classmethod
	def create(cls,fp):
		fp = cls.fp(fp)
		JiviCore.cmd('IF NOT EXIST "{fp}" mkdir "{fp}"'.format(fp=fp))
		return fp
		


class JiviFile:
	@classmethod
	def name(cls,fp,ex=False,lower=False):
		retval = os.path.basename(fp)
		if not ex:
			retval = os.path.splitext(retval)[0]
		
		return retval.lower() if lower else retval

	@classmethod
	def fp(cls,*inp,d=None):
		
		fp_parts 	= inp if d is None else [d] + list(inp)
		fp 			= os.path.join(*fp_parts)
		return os.path.abspath(fp)
	

	@classmethod
	def modified(cls,fp):
		try:
			m = os.path.getmtime(fp)
			return m
		except:
			return 0
	
	@classmethod
	def delete(cls,fp):
		try:
			os.remove(fp)
		except:
			traceback.print_exc()
			return None
	
	@classmethod
	def ex(cls,fp,lower=True):
		if lower: 
			return cls.ex(fp,lower=False).lower()
		try:
			return os.path.splitext(fp)[1][1:]
		except:
			return ""
	
	@classmethod
	def dir(cls,fp):
		return JiviDir.fp(os.path.dirname(fp))
	
	@classmethod
	def exists(cls,fp):
		return os.path.isfile(fp)


class JiviIOReader:
	def __init__(self,io):
		self.io = io
	@property
	def fp(self):
		return self.io.fp
	def all(self,mode='r'):
		try:
			with open(self.fp,mode) as f:
				x = f.read()
				return x
		except:
			traceback.print_exc()

	def all_full(self,encoding='utf-8'):
		try:
			with open(self.fp,'r',encoding=encoding) as f:
				x = f.read()
				x = "\n".join(list(x.replace("\r\n", "\n").splitlines()))
	
				return x
		except:
			if encoding == 'utf-8':
				return self.all_full(fp,encoding='cp1252')
			elif encoding == 'cp1252':
				return self.all_full(fp,encoding='mbcs')
		return None
	
	def lines(self,strip=1,encoding='utf-8'):
		try:
			with open(self.fp,'r',encoding=encoding) as f:
				x = [a.rstrip() if strip else a for a in f.readlines()]
				return x
		except:
			pass
		return []
	
	def lines_stripped(self,fp):
		try:
			with open(self.fp,'r') as f:
				x = [a for a in [a.strip() for a in f.readlines()] if a]
				return x
		except:
			pass
		return []
	


	def lines_special(self,fp):
		
		try:
			x = self.all()
			x = x.splitlines()
			return x
		except:
			pass
	
	
	def jsono(self,encoding='utf-8'):
		try:
			with open(self.fp,"r",encoding=encoding) as f:
				s = f.read()
				a = json.loads(s)
				for k in a.keys():
					return a[k]
		except:
			return None
	
	def json(self,encoding='cp1252'):
		try:
			with open(self.fp,"r",encoding=encoding) as f:
				s = f.read()
				a = json.loads(s)
				return a
		except:
			return None



class JiviIOWriter:
	def __init__(self,io):
		self.io = io
	@property
	def fp(self):
		return self.io.fp
	
	def str(self,s,run=0):
		try:
			with open(self.fp,'w') as f:
				f.write(s)
			if run:
				self.io.run()
		except:
			perr()
			return None
	
	def ar(self,ar,sep='\n'):
		return self.str(sep.join(ar))
	
	def json(self,ob,indent=4,carefull=0,run=0):
		fp = self.fp
		
		def write(to_fp):
			with open(to_fp,'w') as f:
				json.dump(ob,f,indent=indent)
		
		if carefull:
			fp_tmp = self.io.dir + self.io.base + str(time()) + '.tmp'
			write(fp_tmp)
			if self.io.read.json(fp_tmp) is not None:
				shutil.move(fp_tmp,fp)
				return True
			else:
				JiviFile.delete(fp)
			
			return
		
		else:
			write(fp)
		if run:
			self.io.run()
		
	
class JiviIO:
	fp:str
	read:JiviIOReader
	write:JiviIOWriter
	def __init__(self,*a,**b):
		self.fp = JiviFile.fp(*a,**b)
		self.read  = JiviIOReader(self)
		self.write = JiviIOWriter(self)
	
	@property
	def base(self):
		return JiviFile.name(self.fp,ex=True)
	@property
	def baselower(self):
		return JiviFile.name(self.fp,ex=True,lower=True)
	@property
	def name(self):
		return JiviFile.name(self.fp,ex=False)
	@property
	def namelower(self):
		return JiviFile.name(self.fp,ex=False,lower=True)
	@property
	def ex(self):
		return JiviFile.ex(self.fp)
	@property
	def dir(self):
		return JiviFile.dir(self.fp)

	def run(self):
		os.system(f'mf "{self.fp}"')
class JiviStr:

	@classmethod
	def tabs(cls,s,length,sep='\t'):
		tabs        = math.ceil(length/4)
		tabs_string = math.floor(len(s)/4)
		s           = sep*max(0,tabs-tabs_string)
		return s
	
	@classmethod
	def tabbed(cls,s,length,sep='\t'):
		return s + cls.tabs(s,length,sep=sep)
	
		
	@classmethod
	def tab(cls,tab):
		keys    = list(tab.keys())
		max_len = len(max(keys,key=len))
		return "\n".join(["{k} : {v}".format(k=cls.tabbed(k,max_len),v=tab[k]) for k in keys])

	@classmethod
	def minlen(cls,s,i):
		return ''.join([str(s)," "*(i - len(str(s)))])


class JiviAr:
	@classmethod
	def split_by_stepsize(cls,ar:list,stepsize=5,as_gen=True) -> [list]:
		def func():
			for i in range(0, len(ar), stepsize):
				yield ar[i:i + stepsize]

		if not as_gen:
			return list(func())
		return func()
	@classmethod
	def split_into_parts(cls,ar:list,parts=5,as_gen=True) -> [list]:
		def func():
			c = math.ceil(len(ar)/parts)
			if c == 0: return
			for i in range(0, len(ar), c):
				yield ar[i:i + c]
		if not as_gen:
			return list(func())
		return func()

	@classmethod
	def join_start_too(cls,ar:list,sep:str) -> str:
		return sep + sep.join(ar) if len(ar) else ''

class JiVi:
	class ALL:
		JiviCore   = JiviCore
		JiviAr  = JiviAr
		JiviDir = JiviDir
		JiviFile   = JiviFile
		JiviIO  = JiviIO
		JiviStr   = JiviStr
	
	C   = JiviCore
	Ar  = JiviAr
	Dir = JiviDir
	F   = JiviFile
	IO  = JiviIO
	S   = JiviStr

def __test():
	test_ar = ['a1','a2','a3','b1','b2','b3','c1','c2','c3','d1','d2','d3']
	assert len(JiviAr.split_by_stepsize(test_ar,3,as_gen=False)[0]) == 3
	assert len(JiviAr.split_into_parts(test_ar,4,as_gen=False)) == 4
	
if __name__ == "__main__":
	__test()
		
	


{ SIZE | SIZE-CHARS | SIZE-PIXELS } width BY height


ignores = 'new,global,shared,private,protected,public,static,class'.split(',')
VIEW_AS 
  {    combo_box_phrase
     | editor_phrase
     | FILL_IN
          [ NATIVE ]
          [  size_phrase ]
          [ TOOLTIP tooltip ]
    | radio_set_phrase
    | selection_list_phrase
    | slider_phrase
    | TEXT
          [ size_phrase ]
          [ TOOLTIP tooltip ]
    | TOGGLE_BOX
          [ size_phrase ]
          [ TOOLTIP tooltip ]
  }
class phrase:
	starters = []
	@classmethod
	def check(cl,ar,kw=''):
		return kw in cl.starters

class view_as(phrase):
	starters = ['view-as']
	parts = [Part(ors=[combo_box,editor,fill_in,radio_set,selection_list,slider,text,toggle_box],req=1)]
	
class size(phrase):
	starters = ['size','size-chars','size-pixels']
	

DEFINE {[[ NEW [ GLOBAL ]] SHARED ]| 
            [ PRIVATE | PROTECTED | PUBLIC ]
            [ STATIC ][ SERIALIZABLE | NON_SERIALIZABLE ]}
  VARIABLE variable_name
  {{   AS primitive_type_name 
       | AS [ CLASS ]{object_type_name}
       | LIKE field       }[ EXTENT [constant]]} 
  [ SERIALIZE_NAME serialize_name ]
  [ BGCOLOR expression]
  [ COLUMN_LABEL label]
  [ CONTEXT_HELP_ID expression]
  [ DCOLOR expression]
  [ DECIMALS n]
  [ DROP_TARGET ]
  [ FONT expression]
  [ FGCOLOR expression]
  [ FORMAT string]
  [ INITIAL
      { constant |{[ constant[ , constant]...]}}]
  [ LABEL string[ , string]...]
  [ MOUSE_POINTER expression]
  [ NO_UNDO ]
  [[ NOT ] CASE_SENSITIVE ]
  [ PFCOLOR expression]
  {[ view_as_phrase ]}
  {[ trigger_phrase ]}
  
  
  
DEFINE {[[ NEW [ GLOBAL ]] SHARED ]| 
            [ PRIVATE | PROTECTED ][ STATIC ]
            [ SERIALIZABLE | NON_SERIALIZABLE ]}
  TEMP_TABLE temp_table_name[ NO_UNDO ] 
  [ NAMESPACE_URI namespace][ NAMESPACE_PREFIX prefix]
  [ XML_NODE_NAME node_name][ SERIALIZE_NAME serialize_name ] 
  [ REFERENCE_ONLY ]
  [ LIKE table_name
      [ VALIDATE ]
      [ USE_INDEX index_name[ AS PRIMARY ]]...]
  [ LIKE_SEQUENTIAL table_name
      [ VALIDATE ]
      [ USE_INDEX index_name[ AS PRIMARY ]]...]
  [ RCODE_INFORMATION ] 
  [ BEFORE_TABLE before_table_name] 
  [ FIELD field_name
      { AS data_type|  LIKE field[ VALIDATE ]}
  [field_options]
  ]...
  [ INDEX index_name
      [[ AS | IS ][ UNIQUE ][ PRIMARY ][ WORD_INDEX ]]
      {index_field[ ASCENDING | DESCENDING ]}...
  ]...
  
  
view_as = ['view_as',


class combo_box(phrase):
	
  
VIEW_AS COMBO_BOX
  [ LIST_ITEMS item_list | LIST_ITEM_PAIRS item_pair_list]
  [ INNER_LINES lines ] [size_phrase ] [ SORT ]
  [ TOOLTIP tooltip]
  [ SIMPLE | DROP_DOWN | DROP_DOWN_LIST ]
  [ MAX_CHARS characters ] 
  [ AUTO_COMPLETION [ UNIQUE_MATCH ] ]

COMBO_BOX
  [ LIST_ITEMS item_list | LIST_ITEM_PAIRS item_pair_list]
  [ INNER_LINES lines ] [ SORT ]
  [ DROP_DOWN | DROP_DOWN_LIST ]
  [ MAX_CHARS characters ] 
  [ AUTO_COMPLETION [ UNIQUE_MATCH ] ]
  
  
  
EDITOR
  {  size_phrase
      | INNER_CHARS characters INNER_LINES lines
  } 
  [ BUFFER_CHARS chars ]
  [ BUFFER_LINES lines ]
  [ LARGE ]
  [ MAX_CHARS characters ]
  [ NO_BOX ]
  [ NO_WORD_WRAP ]
  [ SCROLLBAR_HORIZONTAL ]
  [ SCROLLBAR_VERTICAL ]
  [ TOOLTIP tooltip ]
  
RADIO_SET
  [ HORIZONTAL [ EXPAND ] | VERTICAL ]
  [ size_phrase ]
  RADIO_BUTTONS label , value [ , label, value ... ] 
  [ TOOLTIP tooltip ]
  
SELECTION_LIST 
  [ SINGLE | MULTIPLE ]
  [ NO_DRAG ]
  [ LIST_ITEMS item_list ]
  [ SCROLLBAR_HORIZONTAL ]
  [ SCROLLBAR_VERTICAL ]
  { size_phrase | INNER_CHARS cols INNER_LINES rows }
  [ SORT ]
  [ TOOLTIP tooltip ]
  
SELECTION_LIST
  [ SINGLE | MULTIPLE ]
  [ NO_DRAG ]
  { LIST_ITEMS item_list| LIST_ITEM_PAIRS item_pair_list}
  [ SCROLLBAR_HORIZONTAL ][ SCROLLBAR_VERTICAL ]
  {size_phrase
    |{ INNER_CHARS cols INNER_LINES rows}
  }
  [ SORT ]
  [ TOOLTIP tooltip]
  

VIEW_AS SLIDER 
  MAX_VALUE max_value MIN_VALUE min_value 
  [ HORIZONTAL | VERTICAL ]
  [ NO_CURRENT_VALUE ]
  [ LARGE_TO_SMALL ]
  [ TIC_MARKS
      { NONE | TOP | BOTTOM | LEFT | RIGHT | BOTH }
      [ FREQUENCY n ]
  ]
  [ TOOLTIP tooltip ]
  [ size_phrase ]


from time import time
import sys
from jivi.fs import *


class Wrapper:
	times = {}
	done_init = 0

	main_setter = File.all(File.fp("main_setter.py",File.dir(os.path.realpath(__file__))))
	
	def error(self,s):
		self.has_error = 1
		try:
			self.before_error()
		except:
			pass
		print(String.minlen(self.name,30) + ' ' + str(s))
	
	def info(self,s):

		print(String.minlen(self.name,20) + String.minlen('info',10) + ' ' + str(s))

	def do(self,todo=0,allinfo=0):
		todo = self.todo if not todo else todo
		mainclass = type(self).mro()[:-1][0]
		if not hasattr(mainclass,"helpertimes"):
			mainclass.helperstarted = time()
			mainclass.helpertimes = {}
		for do in todo:
			if not do in mainclass.helpertimes:
				mainclass.helpertimes[do] = []
		for do in todo:
			ss = time()
			if allinfo:
				self.info("Doing {do}".format(do=do))
			
			if (getattr(self,do)() == False):
				self.error("failed to do {do}".format(do=do))
				return 0
			mainclass.helpertimes[do].append(time() - ss)
			
		if allinfo:
			Wrapper.print_times(mainclass)
		return 1
	
	@staticmethod
	def print_times(mainclass):
		print("\n\n")
		print("printing times for " + mainclass.__name__)
		for k,ar in mainclass.helpertimes.items():
			tsum = sum(ar) * 10000
			tsum = int(tsum)
			t = int(tsum / len(ar))
			tsum = str(tsum)
			t = str(t)
			t = t + (10 - len(t))*" "
			tsum = tsum + (10 - len(tsum))*" "
			
	
			k = k + (50 - len(k))*" "	
			print("{k}  {tsum} {t}".format(k=k,tsum=tsum,t=t))
			
		print(time() - mainclass.helperstarted)
		print("\n\n")


import math
from datetime import datetime
import subprocess

def Popen(s):
	try:
		process = subprocess.Popen(s, stdout=subprocess.PIPE,shell=1)
		stdout, stderr = process.communicate()
		if stderr: print(stderr)
		
		stdout = stdout.decode("cp850")
		for x in stdout.splitlines():
			yield x.strip()
	except:
		pass
	
class to_string:
	__BYTE_HELPER = [[math.pow(1024,i),name] for i,name in enumerate(["bytes","KB","MB","GB","TB"])]
	__BYTE_HELPER.reverse()
	@classmethod
	def bytes(cls,number_of_bytes):
		number_of_bytes = float(number_of_bytes)
		if not number_of_bytes:
			return "0 bytes"
		for a in cls.__BYTE_HELPER:
			n = number_of_bytes/a[0]
			if n >= 1:
				return "{a} {b}".format(a=round(n,2),b=a[1])
	@classmethod
	def sec(cls,s):
		h = math.floor(s/3600)
		s -= h*3600
		m =  math.floor(s/60)
		s -= m*60
		return "%02d:%02d:%02d" % (h,m,s)

	@classmethod
	def unix(cls,i):
		return datetime.utcfromtimestamp(i).strftime('%Y-%m-%d %H:%M:%S')


def to_int(n):
	try:
		retval = int(n)
		return retval
	except:
		return None

		


from __future__ import annotations
import os,json,shutil
from time import time
from typing import List

def inttime():
	return int(time())

get_rel = os.path.relpath

def get_ex_from_fp(fp,lower=False):
	retval =  os.path.splitext(fp)[1][1:]
	return retval.lower() if lower else retval

class JDirChildWrapper:
	def __init__(self,parent:JDirFileWrapper):
		self.parent = parent
	
	@property
	def fp(self):
		return self.parent.fp



class JDirFileWrapper:
	isfile : bool = False
	isdir : bool = False
	__fp : str
	def __init__(self,*parts):
		self.set_fp(*parts)
	
	def set_fp(self,*value):
		self.__fp = os.path.abspath(os.path.join(*value))

	@property
	def all_parent_dir_ob(self):
		p = self.parent_dir_ob
		old_fp = ''
		retval = [] 
		while p.fp != old_fp:
			old_fp = p.fp
			retval.append(p.copyob)
			p = p.copyob.up(1)
		
		return retval
	
	@property
	def parent_dir_ob(self):
		return JDir(self.parent_dir_fp)
	
	@property
	def parent_dir_fp(self):
		if self.isdir:
			return os.path.dirname(os.path.abspath(os.path.join(self.fp,'..')))
		if self.isfile:
			return os.path.dirname(self.fp)
		
		print("HOW SHIT")
		print(self.fp)
		exit()
	
	@parent_dir_fp.setter
	def parent_dir_fp(self,value):
		self.set_fp(value,self.name)

	@property
	def fp(self):
		return self.__fp
	
	@fp.setter
	def fp(self,*value):
		self.set_fp(*value)
	
	@property
	def name(self):
		return os.path.basename(self.fp)
	
	@name.setter
	def name(self,value):
		self.set_fp(self.parent_dir_fp,value)

	@property
	def lname(self):
		return self.name.lower()
	
	@property
	def modified(self):
		return os.path.getmtime(self.fp)
	
	@property
	def created(self):
		return os.path.getctime(self.fp)

	@property
	def copyob(self):
		if self.isdir:
			return JDir(self.fp)
		if self.isfile:
			return JFile(self.fp)
		
		print("copy HOW SHIT")
		print(self.fp)
		exit()

class JFileReader(JDirChildWrapper):
	def json(self,encoding=None,mode='r'):
		try:
			with open(self.fp,mode,encoding=encoding) as f:
				a = json.load(f)
				return a
		except:
			return None
	def text(self,encoding=None,mode='r'):
		try:
			with open(self.fp,mode,encoding=encoding) as f:
				txt = f.read()
				return txt
		except:
			return None
	def lines(self,encoding=None,mode='r',strip=True):

		txt = self.text(encoding=encoding,mode=mode)
		if txt is None:
			return txt
		for x in txt.splitlines():
			if not strip:
				return x
			x = x.strip()
			if x:
				return x
		
class JFileWriter(JDirChildWrapper):
	
	def text(self,txt,mode='w',encoding=None):
		with open(self.fp,mode,encoding=encoding) as f:
			f.write(txt)

	def lines(self,*lines,seperator="\n",mode='w',encoding=None):
		self.text(seperator.join(map(str,lines)))
	
	def json(self,ob,indent=4,carefull=False,debug=False):
		if carefull:
			tmp_file : JFile = self.parent.copyob
			
			tmp_file.name = f"{inttime()}{self.parent.name_without_ex}.tmp"

			if not tmp_file.writer.json(ob,indent=indent,carefull=False):
				if debug:
					print(f"Failed to write {tmp_file}")

				return False
			
			if tmp_file.reader.json() is None:
				if debug:
					print(f"Failed read {tmp_file}")
			else:
				if tmp_file.move(self.fp):
					return True
				else:
					print("Something went wrong")
					exit()
			
			tmp_file.delete()
			return False
			
		else:
			try:

				with open(self.fp,'w') as f:
					json.dump(ob,f,indent=indent)

				return True

			except:
				return False
		
	


class JDirReader(JDirChildWrapper):
	def get(self,asob=True,deep=True,dirs=True,files=True,ex=[],rel=False):
		dirs_todo = [self.fp]
		self_fp = self.fp[:]
		ex_tab = {k.lower() : True for k in ([ex] if isinstance(ex,str) else ex)}
		ex_check = True if ex_tab else False
		rel_path = ''

		while dirs_todo:
			new_dirs_todo = []
			for fp_dir in dirs_todo:
				if rel:
					rel_path = get_rel(fp_dir,self_fp)
					if rel_path == '.':
						rel_path = ''
					else:
						rel_path = rel_path + os.sep
					
				for x in os.listdir(fp_dir):
					fp = fp_dir + os.sep + x
					if os.path.isfile(fp):
						if not files:
							continue
						if ex_check and (not get_ex_from_fp(fp,True) in ex_tab):
							continue
						yield JFile(fp) if asob else (rel_path + x if rel else fp)
						
					else:
						if deep:
							new_dirs_todo.append(fp)
						if not dirs:
							continue
						yield JDir(fp) if asob else (rel_path + x if rel else fp)
			dirs_todo = new_dirs_todo
		
	def files(self,asob=True,deep=True,ex=[],rel=False):
		return self.get(asob=asob,deep=deep,ex=ex,rel=rel,files=True,dirs=False)
	
	def files_ob(self,deep=True,ex=[]) -> List[JFile]:
		return self.get(asob=True,deep=deep,ex=ex,rel=False,files=True,dirs=False)

	def dirs(self,asob=True,deep=True,rel=False):
		return self.get(asob=asob,deep=deep,rel=rel,files=False,dirs=True)

	def dirs_ob(self,deep=True) -> List[JDir]:
		return self.get(asob=True,deep=deep,rel=False,files=False,dirs=True)




class JFile(JDirFileWrapper):
	isfile = True
	isdir = False
	reader : JFileReader
	writer : JFileWriter
	def __init__(self,*parts):
		super().__init__(*parts)
		self.reader = JFileReader(self)
		self.writer = JFileWriter(self)

	def copy(self,fp_to,carefull=False):
		try:
			shutil.copy(self.fp,fp_to)
			retval = JFile(fp_to)
			if carefull and not retval.exists:
				return None
			return retval
		except:
			return None

	def move(self,fp_to):
		try:
			shutil.move(self.fp,fp_to)
			self.fp = fp_to
			return True
		except:
			return False
		
	def start(self,same_window=False,start_dir=None,title=None,wait=False,minimized=False,maximized=False):
		attr = []
		if start_dir:
			attr.append(f'/D "{start_dir}"')
		if minimized:
			attr.append("/MIN")
		if maximized:
			attr.append("/MAX")
		if wait:
			attr.append("/WAIT")
		if same_window:
			attr.append("/B")
		
		attr = ' '.join(attr)
		title = self.name_without_ex if title is None else title
		cmd = f'start {attr} "{self.name_without_ex}" "{self.fp}"'
		print(cmd)
		os.system(cmd)
	
	def delete(self):
		try:
			os.remove(self.fp)
			return True
		except:
			return False
		

	@property
	def exists(self):
		return os.path.isfile(self.fp)

	def __repr__(self):
		return f'[JFile] {self.fp}'
	
	@property
	def ex(self):
		return os.path.splitext(self.fp)[1][1:]
	
	@ex.setter
	def ex(self,value):
		self.name = self.name_without_ex + '.' + value
	

	@property
	def name_without_ex(self):
		return os.path.splitext(os.path.basename(self.fp))[0]
	
	@property
	def lname_without_ex(self):
		return self.name_without_ex.lower()
	
	@property
	def lex(self):
		return self.ex.lower()
	
	@property
	def size(self):
		return os.path.getsize(self.fp)
	
class JDir(JDirFileWrapper):
	isfile = False
	isdir = True
	reader : JDirReader
	def __init__(self,*parts):
		super().__init__(*parts)
		self.reader = JDirReader(self)


	def isEmpty(self):
		for x in os.listdir(self.fp):
			return False
		return True


	@classmethod
	def cwd(cls):
		return cls(os.getcwd())
	

	def file(self,*parts) -> JFile:
		return JFile(self.fp,*parts)
	
	def up(self,n=1):
		self.set_fp(self.fp,*['..']*n)
		return self
	
	def down(self,*parts):
		self.set_fp(self.fp,*parts)
		return self

	def rel(self,fpOrDir):
		fp = fpOrDir.fp if isinstance(fpOrDir,JDir) else fpOrDir
		return os.path.relpath(self.fp,fp)

	def __repr__(self):
		return f'[JDir] {self.fp}'

	@property
	def exists(self):
		return os.path.isdir(self.fp)
	
	def create(self):
		if not self.exists:
			os.makedirs(self.fp)
		return self.exists

	
	def delete_full(self):
		try:
			if self.exists:
				shutil.rmtree(self.fp)

			return not self.exists
		except:
			return None

	def delete(self):
		try:
			os.rmdir(self.fp)
			return True
		except:
			return False




from Imports import *

from Main import Main
from Compiler import Compiler
#Compiler.todo = Compiler.todoss

exec(Wrapper.main_setter)
M = Main()



set_main(M)
c = Compiler(M,r'c:\\develop\\l_intex\\framework\\prg\\mainimag.w').do()

exit()
for fp in SESSION.all_files:
	x = File.all(SESSION.fp_compiled(fp,1))
	if x.find("{&") != -1:
		print(SESSION.fp_compiled(fp,1))

		
		
exit()

for fp in SESSION.all_files:
	
	if File.ex(fp).lower() == "w":
		c = Compiler(M,fp)
		c.do()
		exit()


from __future__ import annotations
from jivi import jvWrapper
import fnmatch,curses, math
from time import time
from typing import List,Dict

@jvWrapper()
class CoordinateStartStop:
	def __init__(self,start:int,stop:int):
		self.start = start
		self.stop  = stop
		self.delta = stop - start

@jvWrapper()
class Coordinate:
	def __init__(self,x:int,y:int):
		self.x = x
		self.y = y

@jvWrapper()
class MouseClick:
	def __init__(self,x:int,y:int):
		self.x = x
		self.y = y
		self.t = time()

class LineResponse:
	def __init__(self,start:Coordinate,stop:Coordinate):
		self.start = start
		self.stop  = stop
	def __repr__(self):
		return f'{self.start} -> {self.stop}'

@jvWrapper()
class InputChar:
	code:int
	c:str
	name:str
	def __init__(self,code:int,name:str=None):
		self.code = code
		self.c = chr(code)
		self.name = name
	
	def __repr__(self):
		return f'Code : {self.code} c : {self.c} name : {self.name}'

@jvWrapper(propnames_to_skip=['item'])
class MenuItemToDisplay:
	x_data : CoordinateStartStop
	line_number = -1
	def __init__(self,row_number:int,result_number:int,selected:bool,item:MenuItem):
		self.row_number = row_number
		self.result_number = result_number
		self.selected = selected
		self.item = item
		self.reset()

	def is_clicked(self,mouse_click:Coordinate):
		if self.line_number != mouse_click.y:
			return False
		#zou nog moeten verbeterd worden....
		return True

	def reset(self):
		self.line_number = -1
		self.x_data : CoordinateStartStop = CoordinateStartStop(-1,-1)
	
	def writeLine(self,C:Curses):
		self.reset()
		ret_val = C.line(self.item.name,selected=self.selected,prefix=f'{str(self.result_number).zfill(2)} {str(self.row_number).zfill(2)}   ')
	
		if ret_val:
			self.line_number = ret_val.start.y
			self.x_data = CoordinateStartStop(ret_val.start.x,ret_val.stop.x)
		return True

@jvWrapper(propnames_to_skip=['func'])
class MenuItem:
	def __init__(self,name,func=None,desc=None,cmd_keys=[],exit_after:bool=False,data=None):
		self.name 			:str		= name
		self.func 						= func
		self.desc 			:str		= desc if desc is not None else name
		self.cmd_keys	  	:List[str]  = cmd_keys
		self.exit_after 				= exit_after
		self.data						= data
	def exec(self,p):
		if not self.func:
			print(f'{self}.exec func is none')
			return
		p.itemClicked = self
		retval = self.func(p)
		# misschien fout hier ?!
		p.itemClicked = None
		if self.exit_after:
			exit()
		return retval

class _SearchFilterHelper:
	@classmethod
	def fnmatch(cls):
		def ret_val(str_to_check:str,looking_for:str):
			return fnmatch.fnmatch(str_to_check,looking_for)
		return ret_val
	@classmethod
	def default(cls):
		def ret_val(str_to_check:str,looking_for:str):
			return fnmatch.fnmatch(str_to_check,'*' + looking_for + '*')
		return ret_val

class SearchFilter:
	fnmatch = _SearchFilterHelper.fnmatch()
	default = _SearchFilterHelper.default()

def _getcurses_codes():
	codeToName = {int(v) : k.lower()[4:] for k,v in curses.__dict__.items() if k.lower().startswith('key_')}
	extra_keys = dict(
		enter=[10,459],
		esc=[27],
		backspace=[8],
		tab=[9]
	)
	for keyname,codes in extra_keys.items():
		for code in codes:
			codeToName[code] = keyname
	nameToCode = dict()
	for k,v in codeToName.items():
		if not v in nameToCode:
			nameToCode[v] = []
		nameToCode[v].append(k)
	return codeToName,nameToCode
codeToName, nameToCode = _getcurses_codes()

class Curses:
	def __init__(self,stdscr):
		self.stdscr : curses.window = stdscr
		curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
		curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_CYAN)
		self.colors_normal = curses.color_pair(1)
		self.colors_selected = curses.color_pair(2)
	
	@classmethod
	def keyname(cls,keyname,quit=False):
		if isinstance(keyname,int):
			if keyname in codeToName:
				return codeToName[keyname]
			keyname = chr(keyname)
		if keyname in nameToCode:
			return keyname
		if len(keyname) == 1:
			code = ord(keyname)
			if code in codeToName:
				return codeToName[code]
		if quit:
			print(f'Invalid keyname {keyname}')
			exit()
		return None
	
	def clear(self):
		self.stdscr.erase()
	
	@property
	def cursor_position(self) ->Coordinate:
		y,x = self.stdscr.getyx()
		return Coordinate(x,y)
	
	def emptyLine(self,n=1) ->LineResponse:
		start = self.cursor_position
		self.stdscr.addstr("\n"*n)
		return LineResponse(start,self.cursor_position)
	
	def line(self,txt:str='',selected=False,color=None,underline=False,bold=False,prefix='',add_line_break=True) ->LineResponse:
		start = self.cursor_position
		color = (self.colors_selected if selected else self.colors_normal) if color is None else color
		attr = color
		for should_do,attrValue in [
			[underline,curses.A_UNDERLINE],
			[bold,curses.A_BOLD]]:
			if should_do:
				attr = attr | attrValue
		if prefix:
			self.stdscr.addstr(prefix)
		self.stdscr.addstr(txt,attr)
		stop = self.cursor_position
		if add_line_break:
			self.stdscr.addstr("\n")
		return LineResponse(start,stop)
	
	def get_char(self) -> InputChar:
		code = self.stdscr.getch()
		return InputChar(code=code,name=self.keyname(code))

@jvWrapper(propnames_to_skip=['menu','c','items','itemClicked'])
class Param:
	inputChar   : InputChar         
	menu        : Menu              
	c           : Curses            
	items       : MenuItemCollection
	itemClicked : MenuItem          
	def __init__(self,menu:Menu,title:str,max_rows:int,max_rows_auto:bool,max_amount_rows_not_data:int):
		self.menu                                          = menu
		self.title                                         = title
		self.max_rows                                      = max_rows
		self.max_rows_auto                                 = max_rows_auto
		self.max_amount_rows_not_data                      = max_amount_rows_not_data
		self.items                    : MenuItemCollection = None
		self.itemClicked              : MenuItem           = None
		self.mouseClick               : Coordinate         = None
	
class MenuItemCollectionIndexHelper:
	def __init__(self,p:Param):
		self.p = p
		self.__selected = 0
	
	@property
	def filtered(self):				return self.p.items.filtered

	@property
	def first(self):				return max(0,self.first_item)

	@property
	def last(self):					return min(len(self.filtered) - 1, self.first + self.p.max_rows - 1)

	@property
	def selected(self):				return self.__selected

	@selected.setter
	def selected(self,new_value):
		last_value           = self.__selected
		new_value            = min(max(0,new_value),len(self.filtered)-1)
		delta_selected_index = (new_value-last_value)

		if delta_selected_index == 0:
			return
		elif delta_selected_index == 1:
			#one down
			if (self.last == last_value):
				self.first_item += 1
		elif delta_selected_index == -1:
			#one up
			if (self.first == last_value):
				self.first_item -= 1
		else:
			self.first_item = min(max(0,new_value - math.floor(self.p.max_rows/2)),max(0,len(self.filtered) - self.p.max_rows))
		self.__selected = new_value
	
	def clear(self):
		self.__selected = 0
		self.first_item = 0


@jvWrapper(propnames_to_skip=['p'])
class MenuItemCollection(list):
	to_display : List[MenuItemToDisplay]
	def __init__(self,p:Param, *items:List[MenuItem]):
		self.filtered : List[MenuItem] = []
		self.p = p
		self.p.items = self
		self.filters = []
		self.index = MenuItemCollectionIndexHelper(p)
		self.clear()
		self.append(*items)


		@self.menu.event.mouse_clicked()
		def mouse_clicked(p:Param):
			mouse_click = p.mouseClick
			def get_menuItem_clicked():
				for a in self.to_display:
					if a.is_clicked(mouse_click):
						return a
			menuItemClicked = get_menuItem_clicked()
			if not menuItemClicked:
				return True
			self.index.selected = menuItemClicked.result_number
			self.menu.key.do('enter')
			return True
	
	@jvWrapper()
	def clear(self):
		self.to_display = []
		self.index.clear()
		self.cmd_keys = dict()
		super().clear()
	
	def __check_item(self,item:MenuItem):
		for sc in item.cmd_keys:
			if sc in self.cmd_keys:
				print(f'{sc} {item} already in cmd_keys !')
				continue
			self.cmd_keys[sc] = item
		return True
	
	def append(self,*ar:List[MenuItem]):
		for v in filter(self.__check_item,ar):
			super().append(v)

	
	def __setitem__(self, key, v:MenuItem):
		return super().__setitem__(key, v) if self.__check_item(v) else None
	
	def set_filtered(self) ->List[MenuItem]:
		def check_all_filters(item:MenuItem):
			for filter_function in self.filters:
				if not filter_function(self.p,item):
					return False
			return True
		self.filtered  = list(filter(check_all_filters,self))
		self.p.menu.event_call.items_to_print_loaded()
		return self.filtered
	
	@property
	def selected(self) -> MenuItem:
		try:
			x = self.filtered[self.index.selected]
			return x
		except:
			print(f'Failed so select menuitem len filtered {len(self.filtered)} index : {self.index.selected}')
			return None
		
	def get_to_display(self) -> List[MenuItemToDisplay]:
		self.set_filtered()
		def get_ret_val():
			row_number = 0
			for result_number in range(self.index.first,self.index.last+1):
				row_number += 1
				selected = result_number == self.index.selected
				item = self.filtered[result_number]
				yield MenuItemToDisplay(row_number=row_number,result_number=result_number,selected=selected,item=item)
		self.to_display = list(get_ret_val())
		return self.to_display
	
	@property
	def menu(self) -> Menu:
		return self.p.menu


@jvWrapper(propnames_to_skip=['p'])
class KeyboardEvent:
	def __init__(self,p:Param):
		self.p = p
		self.__tab = dict()
	def __call__(self, keyname):
		keyname = Curses.keyname(keyname,quit=True)
		def decorator(f):
			self.__tab[keyname] = f
			return f
		return decorator
	def do(self,keyname):
		keyname = Curses.keyname(keyname)
		if not keyname:
			return
		if keyname in self.__tab:
			self.__tab[keyname](self.p)

def EventHelper_inner_function():
	def main_decorator(mainf):
		name = mainf.__name__
		def ret_val(self,*a,**b):
			def add_function(func_to_add):
				if not name in self._funcs:
					self._funcs[name] = []
				self._funcs[name].append(func_to_add)
			
			def special_modus():
				if self._modus == 'add':
					add_function(a[0])
					return
				if self._modus == 'call':
					if not name in self._funcs:
						return True
					for func in self._funcs[name]:
						if not func(self.p):
							return False
					return True
					

			if self._modus != 'normal':
				ret_val = special_modus()
				self._modus = 'normal'
				return ret_val
				
			def decorator(f):
				add_function(f)
				return f
			
			return decorator
		return ret_val
	return main_decorator


def EventHelper_inner_function_strontboel():
	def main_decorator(mainf):
		name = mainf.__name__
		def ret_val(self,*a,**b):
			def add_function(func_to_add):
				if not name in self._funcs:	
					self._funcs[name] = []
				self._funcs[name].append(func_to_add)
				return func_to_add

			def decorator(f):
				return add_function(f)
			
			if self._modus == 'normal':		
				return decorator

			if self._modus == 'add':		
				return add_function(a[0])
			
			if self._modus == 'call':
				for func in getattr(self._funcs,name,[]):
					if not func(self.p):
						return False
				return True
			return None
		return ret_val
	return main_decorator

@jvWrapper(propnames_to_skip=['p'])
class MenuEvent:
	_modus  = 'normal'
	__event = EventHelper_inner_function
	def __init__(self,p:Param):
		self._modus = 'normal'
		self.p = p
		self._funcs = dict()
	
	@__event()
	def mouse_clicked(self):			pass
	@__event()
	def curses_started(self):			pass
	@__event()
	def items_to_print_loaded(self):	pass
	@__event()
	def before_menu_build(self):		pass
	@__event()
	def after_menu_build(self):			pass
	@__event()
	def char_received(self):			pass
	@__event()
	def after_items_print(self):		pass

	def _set_modus(self,modus):
		self._modus = modus
		return self
	
	@property
	def add(self):
		return self._set_modus('add')
	
	@property
	def call(self):
		return self._set_modus('call')
	
	@property
	def info(self):
		return self._set_modus('info')


@jvWrapper(propnames_to_skip=['p'])
class MenuStartStringSelect:
	def __init__(self,p:Param):
		self.p = p
		self.enabled = False
		self.input = ''
		self.time_last_added = 0
		self._load_key_listeners()
	
	@property
	def menu(self) -> Menu:
		return self.p.menu
	
	def _load_key_listeners(self):
		key = self.menu.key
		events = self.menu.event
		@events.items_to_print_loaded()
		def items_to_print_loaded(p:Param):
			p.items.filtered.sort(key = lambda x : x.name.lower())
			if not self.enabled:
				return True
			if not self.input:
				return True
			def select_index():
				for i,c in enumerate(p.items.filtered):
					if c.name.lower().startswith(self.input):
						p.items.index.selected = i
						return
			select_index()
			return True
		@events.char_received()
		def input_checker(p:Param):
			inp = p.inputChar
			def valid():
				if p.menu.search.searching:
					print('searching')
					return True
				if inp.name:
					return True
				if not inp.c:
					return True
			if valid():
				self.disable()
				return True
			self.add_char(inp.c)
			return True
	
	def disable(self):
		self.enabled = False
		self.input = ''
		self.time_last_added = 0
	
	def add_char(self,c:str):
		self.enabled = True
		time_now = time()
		if (time_now - self.time_last_added) > 1:
			self.input = ''
		self.time_last_added = time_now
		self.input = self.input + c
		self.input = self.input.lower()
		print(f'input : {self.input}')

@jvWrapper(propnames_to_skip=['p'])
class MouseListener:
	def __init__(self,p:Param):
		self._funcs = dict()
		self.p = p
		self._load_key_listeners()
		
	@property
	def menu(self) -> Menu:
		return self.p.menu
	
	@property
	def mouse_coords(self) -> Coordinate:
		_, mx, my, _, _ = curses.getmouse()
		return Coordinate(mx,my)
	
	def _load_key_listeners(self):
		key = self.menu.key
		events = self.menu.event
		@events.curses_started()
		def curses_started(p:Param):
			curses.mousemask(1)			
			return True
		@key(curses.KEY_MOUSE)
		def on_mouse_click(p:Param):
			p.mouseClick = self.mouse_coords
			p.menu.event_call.mouse_clicked()
			return True
	
@jvWrapper(propnames_to_skip=['p'])
class MenuSearch:
	def __init__(self,p:Param):
		self.p = p
		self.searching = False
		self.search_input = ''
		self._load_key_listeners()
	
	@property
	def menu(self) -> Menu:
		return self.p.menu
	
	def _load_key_listeners(self):
		key = self.menu.key

		@key(curses.KEY_BACKSPACE)
		def back_space(p:Param):
			if self.searching:
				self.search_input = self.search_input[:-1]
			return True
		
		@key('f2')
		def search(p:Param):
			self.searching = not self.searching
			print(f'Searching : {self.searching}')
			self.search_input = ''
		event : MenuEvent = self.p.menu.event

		@event.after_items_print()
		def after_items_print(p:Param):
			print('after_items_print_listener')
			if not self.searching:
				return True
			C = p.c
			C.emptyLine(2)
			C.line(f'#results : {len(self.p.items.filtered)}/{len(self.p.items)}')
			C.line(f'Searching : {self.search_input}',add_line_break=False)
			return True
		
		@event.char_received()
		def char_received(p:Param):
			inp = p.inputChar
			if self.searching and (inp.c) and (not inp.name):
				self.search_input = self.search_input + inp.c
			return True

		
		def search_filter(p:Param,item:MenuItem):
			if not (self.searching and self.search_input):
				return True
			return SearchFilter.default(item.name,self.search_input)
		self.menu.items.filters.append(search_filter)

class Menu:
	def __init__(self,*menuItems : List[MenuItem],start=False,title='',max_rows=25,max_rows_auto=False,max_amount_rows_not_data=7):
		self.p                 = Param(self,title=title,max_rows=max_rows,max_rows_auto=max_rows_auto,max_amount_rows_not_data=7)
		self.key               = KeyboardEvent(self.p)
		self.event             = MenuEvent(self.p)
		self.items             = MenuItemCollection(self.p,*menuItems)
		self.search            = MenuSearch(self.p)
		self.startStringSelect = MenuStartStringSelect(self.p)
		self.mouseListener     = MouseListener(self.p)
		self._load_key_listeners()
	
	def _load_key_listeners(self):
		key = self.key
		@key('esc')
		def f(p:Param):		exit()
		@key('up')
		def f(p:Param):		p.items.index.selected -= 1
		@key('down')
		def f(p:Param):		p.items.index.selected += 1
		@key('enter')
		def f(p:Param):		p.items.selected.exec(p)
		@key('ppage')
		def f(p:Param):		p.items.index.selected -= p.max_rows
		@key('npage')
		def f(p:Param):		p.items.index.selected += p.max_rows
	
		@self.event.curses_started()
		def curses_started(p:Param):
			p.max_rows = curses.LINES - p.max_amount_rows_not_data if p.max_rows_auto else p.max_rows
			return True
	
	@property
	def event_add(self):		return self.event.add

	@property
	def event_call(self):		return self.event.call
	
	def item(self,name,desc=None,cmd_keys=[],exit_after=False,data=None):
		def decorator(f):
			self.items.append(MenuItem(name,f,desc=desc,cmd_keys=cmd_keys,exit_after=exit_after,data=data))
			return f
		return decorator
	
	def display(self):
		def curses_wrapper(stdscr):

			C = Curses(stdscr)
			self.p.c = C
			self.event_call.curses_started()
			def printer():
				if not self.event_call.before_menu_build():
					print(f'before_menu_build failed')
					return False
				
				C.clear()
				if self.p.title:
					C.line(self.p.title,underline=True)
					C.emptyLine(2)
				


				for displayItem in self.items.get_to_display():
					
					displayItem.writeLine(C)
				
				if not self.event_call.after_items_print():
					print(f'after_items_print failed')
					return False
				
				inputChar = C.get_char()
				self.p.inputChar = inputChar
				if not self.event_call.char_received():
					print(f'char_received failed')
					return False
				
				self.key.do(inputChar.code)
				if not self.event_call.after_menu_build():
					print(f'after_menu_build failed')
					return False
			
			while True:
				if printer() == False:
					exit()


		curses.wrapper(curses_wrapper)


from jivi.fs import *
x = File.jread(r'C:\tools\py64\Lib\site-packages\jivi\oe\File\replaced.txt')
print(x[403])
x = File.jread(r'C:\tools\py64\Lib\site-packages\jivi\oe\File\commands.txt')

ind = 0
for a in x:
	for c in a:
		if isinstance(c,int):
			if c != ind:
				print(c)
				exit()
			ind += 1


from .internal import DownloadFileInternal,to_string
from time import time
from jivi import jvWrapper

class _DownloadHelperData:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	@property
	def total(self):						return self.__downloader.size
	
	@property
	def downloaded(self):					return self.__downloader.dblock_bytes_loaded

	@property
	def written(self):						return self.__downloader.writer_bytes_written

	@property
	def to_download(self):					return self.total - self.downloaded

	@property
	def to_write(self):						return self.total - self.written

	@property
	def speed(self):						return self.__downloader.info_download_speed



class _DownloadHelperDataHuman:
	__data : _DownloadHelperData
	def __init__(self,data):
		self.__data = data

	@property
	def total(self) -> str:				return to_string.bytes(self.__data.total)
	
	@property
	def downloaded(self) -> str:		return to_string.bytes(self.__data.downloaded)
	
	@property
	def written(self) -> str:			return to_string.bytes(self.__data.written)
	
	@property
	def to_download(self) -> str:		return to_string.bytes(self.__data.to_download)
	
	@property
	def to_write(self) -> str:			return to_string.bytes(self.__data.to_write)
	
	@property
	def speed(self) -> str:				return to_string.bytes(self.__data.speed) + '/s'



class _DownloadHelperTime:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader
	
	@property
	def remaining(self):
		speed = self.__downloader.info.speed
		to_download = self.__downloader.src.size - self.__downloader.blocks.bytes_downloaded
		if speed == 0:
			return -1
		return to_download / speed


	@property
	def elapsed(self):				return time() - self.__downloader.time_started

	@property
	def started(self):				return self.__downloader.time_started

	@property
	def ended(self):				return self.__downloader.time_ended



class _DownloadHelperTimeHuman:

	__time : _DownloadHelperTime
	def __init__(self,timeOb):
		self.__time = timeOb

	@property
	def remaining(self) -> str:		return to_string.sec(self.__time.remaining)
	
	@property
	def elapsed(self) -> str:		return to_string.sec(self.__time.elapsed)
	
	@property
	def started(self) -> str:		return to_string.unix(self.__time.started)
	
	@property
	def ended(self) -> str:			return to_string.unix(self.__time.ended)

class _DownloadHelperError:
	src          : str                 
	download     : str                 
	write        : str                 
	__downloader : DownloadFileInternal
	
	def __init__(self,downloader):
		self.__downloader = downloader
	@property
	def has(self):
		return self.src or self.download or self.write
	
class _DownloadHelperStatus:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	@property
	def finished(self) -> bool:				return self.__downloader.status_finished
	
	@property
	def running(self) -> bool:				return self.__downloader.status_running

	@property
	def stopped(self) -> bool:				return not self.__downloader.status_running

	@property
	def started(self) -> bool:				return self.__downloader.status_finished or self.__downloader.status_running

	@property
	def error(self) -> bool:
		return False
#		return self.__downloader.has_errors
	

class _DownloadHelperControl:
	__downloader : DownloadFileInternal
	def __init__(self,downloader):
		self.__downloader = downloader

	def start(self):							return self.__downloader.start()
	def stop(self):								return self.__downloader.stop()
	def reset(self):							return self.__downloader.reset()
	def move(self,new_fp=r'C:\temp\temp.tmp'):	return self.__downloader.move(new_fp)



class Download:
	__downloader : DownloadFileInternal              
	data         : _DownloadHelperData     
	dataHuman    : _DownloadHelperDataHuman
	time         : _DownloadHelperTime     
	timeHuman    : _DownloadHelperTimeHuman
	error        : _DownloadHelperError    
	status       : _DownloadHelperStatus   
	control      : _DownloadHelperControl  

	def __init__(self,url:str,fp:str, block_size_kb : int = 400, max_threads : int = 10):
		self.__downloader = DownloadFileInternal(url,fp,block_size_kb,max_threads)
		self.data         = _DownloadHelperData(self.__downloader)
		self.dataHuman    = _DownloadHelperDataHuman(self.data)
		self.time         = _DownloadHelperTime(self.__downloader)
		self.timeHuman    = _DownloadHelperTimeHuman(self.time)
		self.error        = _DownloadHelperError(self.__downloader)
		self.status       = _DownloadHelperStatus(self.__downloader)
		self.control 	  = _DownloadHelperControl(self.__downloader)



from jivi.oe.Imports import *
exec(Wrapper.main_setter)

import configparser,os,win32com.client,datetime
from jivi.oe.Compiler import Compiler
from jivi.oe.Session import Session


shortcut_link = r"C:\Users\jorisv\Desktop\Intex_links\local\MEC.lnk"






class Env(Wrapper):
	name = "Env"
	todo = ['read_link','set_arg','set_sections']
	file_types=['ininame','pf']
	sections=['Startup','Colors','fonts']
	reload = Data(
					env = 0
				)

	def __init__(self,pMAIN,link=shortcut_link):
		set_main(pMAIN)


		
		FP.p_database_structure = File.fp('p_database_structure.p',FP.static)
		self.all_files = {}
		self.link = File.fp(link)
		self.global_scopes = {}
		self.do(allinfo=0)
		self.info('loaded')

	def read_link(self):
		shell = win32com.client.Dispatch("WScript.Shell")
		self.shortcut = shell.CreateShortCut(self.link)
		self.path = self.shortcut.Targetpath
		self.wd = self.shortcut.WorkingDirectory
		return 1

	def set_arg(self):
		self.arg = {}
		arg = [a for a in [a.strip() for a in self.shortcut.Arguments.split(' ')] if a]
		meep = []
		key = 0
		for a in arg:
			if a.startswith('-'):
				if key:
					self.arg[key] = ' '.join(meep)
					meep = []
				key = a[1:]
			elif key:
				meep.append(a)
		if key:
			self.arg[key] = ' '.join(meep)
		for k,v in self.arg.items():
			if not hasattr(self,k):
				setattr(self,k,v)
		for a in self.file_types:
			if a in self.arg:
				if not os.path.isfile(self.arg[a]):
					self.arg[a] = os.path.join(self.wd,self.arg[a])

	def set_sections(self):
		self.propath = []
		config = configparser.ConfigParser(strict=0)
		config.read(self.arg['ininame'])
		for a in self.sections:
			t = {}
			sn = a.lower()
			if a in config:
				ca = config[a]
				for x in ca:
					t[x.lower()] = ca[x]
			setattr(self,sn,t)
		for a in self.startup['propath'].split(','):
			fp = self.fp_dir(a.strip())
			if fp:
				self.propath.append(fp.lower())

	def fp_dir(self,fp):
		if not os.path.isdir(fp):
			fp = os.path.join(self.wd,fp)
		if not os.path.isdir(fp):
			return 0
		return os.path.abspath(fp).rstrip(os.sep) + os.sep
	def run_reload_env(self):
		new_args = {k : v for k,v in self.arg.items()}
		new_args['p'] = FP.p_database_structure
		new_args['T'] = FP.temp
		new_args = ["-{k} {v}".format(k=k,v=v) for k,v in new_args.items()]
		
		os.system("{path} {new_args}".format(path=self.path,new_args=" ".join(new_args)))
		
	def load_tabs(self):
		files = [(name,FP.temp + name + ".json") for name in ['ttField','ttIndex','ttIndexField','ttFile','ttFont','ttSession']]
		self.tabs = {}
		for name,fp in files:
			t = File.joread(fp)
			if t == None: return None
			self.tabs[name[2:].lower()] = t
			
		return 1
	def load_env(self):
		if not self.reload.env:
			if self.load_tabs(): return 1
		self.run_reload_env()
		return self.load_tabs()


		


from jivi.oe.Imports import *
exec(Wrapper.main_setter)


from jivi.oe.Env import Env
from jivi.oe.Session import Session
from jivi.oe.Compiler import Compiler





def log_traceback(ex, ex_traceback=None):
	os.system("\n")
	if ex_traceback is None:
		ex_traceback = ex.__traceback__
	tb_lines = [ line.rstrip('\n') for line in traceback.format_exception(ex.__class__, ex, ex_traceback)]
	print("\n".join(tb_lines))





		
class FPC(Wrapper):
	def __init__(self,MAIN):
		set_main(MAIN)
		self.main = File.dir(os.path.realpath(__file__))
		self.temp = Dir.create(self.main + '_temp')
		self.compiled_first_step = Dir.create(self.main + '_compiled_first_step')
		self.compiled = Dir.create(self.main + '_compiled')
		self.static = Dir.create(self.main + "_static")
		self.data = Dir.create(self.main + '_data')
		
		self.test = r'C:\develop\l_intex\Intex\incl\intex_maintrig_inq_lookup.i' #self.static + 'test.p'
		
class Main:
	FP = 0
	ENV = 0
	SESSION = 0
	REPLACER = Data()
	def __init__(self):
		self.FP = FPC(self)
		self.ENV = Env(self)
		self.SESSION = Session(self)
		self.todocounter = 0
		set_main(self)
		
	def compile_all_helper(self,ar):
		for fp in ar:
			c = Compiler(self,fp)
			c.do()
		self.todocounter -= 1
	def compile_all(self):
		STEPS = 5
		self.todocounter = STEPS
		for ar in Ar.steps(SESSION.all_files,STEPS):
			Thread(self.compile_all_helper,(ar,))
		
		while self.todocounter:
			sleep(1)
		
		
		Wrapper.print_times(Compiler)

	def test(self):
		c = Compiler(self,txt=File.all(FP.test))

	def compile(self,fps):
		for i,fp in enumerate(fps):
			print(fp + "\n")
			try:
				
				c = Compiler(self,fp=fp)
				if not c.do():
					print("failed to compile " + fp)
					c.write_ar_readable()
					exit()
			except Exception as ex:
				print("Main compile error !")
				print(fp)

				log_traceback(ex)
				print("yay")
				return (0,i,fp)
		return (1,0,0)
if __name__ == "__main__":
	MAIN = Main()

	for a in sys.argv[1:]:
		a = a.lower().strip()
		if a in ['ca']:
			MAIN.compile_all()
		elif a in ['t']:
			MAIN.test()


import setuptools
from jivi.fs import JFile,JDir
from jivi.wrapper import jvWrapper,jvWrapperSuper

@jvWrapper()
class HelperClass(jvWrapperSuper):
	def __init__(self):
		def find_settings_dir() -> JDir:
			for d in JFile(__file__).all_parent_dir_ob:
				dSetting = d.copyob.down('settings')
				if dSetting.exists:
					return dSetting
			

		self.dSettings = find_settings_dir()
		settings_fp = '.'
		if not self.dSettings:
			print("Settings dir not found")
			exit()
		else:
			settings_fp = self.dSettings.fp

		
		self.config           = JFile(self.dSettings.fp,'config.json').reader.json()
		self.long_description = JFile(self.dSettings.fp,"README.md").reader.text()
		self.version          = self.config.get('version')


	def info(self):
		self.jvlog.info('info')

Helper = HelperClass()

setuptools.setup(
    name="jivi",
    version=Helper.version,
    author="Joris Vercleyen",
    author_email="jorisvercleyen@gmail.com",
    description="Joris Vercleyen his personal library",
    long_description=Helper.long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


from . import JDir,JFile

from .. import util

fp_test_dir = "C:\\temp\\testdir"


class FSTester(util.Tester):
	def files(self):
		def create_remove():
			d = JFile(fp_test_dir,"testfile.txt")
			d.delete()
			if d.exists:
				return "delete"
				
			d.writer.text("test")
			if not d.exists:
				return "write create"

			txt = d.reader.text()
			if txt is None:
				return "reader"

			if txt != "test":
				return "reader wrong text"
		

		def json():
			d = JFile(fp_test_dir,"testfile.json")
			test_ob = dict(a=1,b=4)
			if not d.writer.json(test_ob,carefull=True):
				return "json"

			d.delete()
			if d.exists:
				return "delete"
		return [create_remove,json]

	
	def dirs(self):
		import os


		def create_remove():
			d = JDir(fp_test_dir)
			
			d.delete_full()
			if d.exists:
				return "delete"
				
			JDir(fp_test_dir).create()
			if not d.exists:
				return "create"
		
		return [create_remove]
	
		







